#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..vcstool import VcsTool
from .bashtoolmixin import BashToolMixIn


class hgTool(BashToolMixIn, VcsTool):
    """Mercurial SCM.

Home: https://www.mercurial-scm.org/
"""
    __slots__ = ()

    def getDeps(self):
        return ['bash']

    def _installTool(self, env):
        self._install.debrpm(['mercurial'])
        self._install.emerge(['dev-vcs/mercurial'])
        self._install.pacman(['mercurial'])
        self._install.apk(['mercurial'])
        self._install.brew('mercurial')

    def autoDetectFiles(self):
        return '.hg'

    def _getCurrentBranch(self, config):
        return self._executil.callExternal([
            config['env']['hgBin'], 'branch'
        ]).strip()

    def vcsGetRepo(self, config, wc_dir=None):
        wc_dir = wc_dir or self._os.getcwd()
        return self._executil.callExternal([
            config['env']['hgBin'], '--repository', wc_dir, 'paths', 'default'
        ], verbose=False).strip()

    def _hgCheckoutTool(self, config):
        hgBin = config['env']['hgBin']
        help_res = self._executil.callExternal(
            [hgBin, 'checkout', '--help'], verbose=False)
        tool_args = []

        if help_res.find('--tool') > 0:
            tool_args += ['--tool', 'internal:merge']

        return tool_args

    def vcsCheckout(self, config, vcs_ref):
        ospath = self._ospath
        hgBin = config['env']['hgBin']
        wc_dir = self._os.getcwd()

        if ospath.isdir('.hg'):
            if 'vcsRepo' in config:
                remote_info = self.vcsGetRepo(config, '.')
                if remote_info != config['vcsRepo']:
                    self._errorExit("Hg remote mismatch: " + remote_info)

            self._executil.callExternal([hgBin, 'pull'])
        else:
            self._executil.callExternal(
                [hgBin, 'clone', config['vcsRepo'], wc_dir])

        if not vcs_ref:
            # skip default branch
            return

        for v in self._executil.callExternal([hgBin, 'branches'], verbose=False).strip().split("\n"):
            if v and v.split()[0] == vcs_ref:
                break
        else:
            for v in self._executil.callExternal([hgBin, 'tags'], verbose=False).strip().split("\n"):
                if v and v.split()[0] == vcs_ref:
                    break
            else:
                self._errorExit(
                    'Unknown VCS ref {0}. Hint: closed branches are ignored!'.format(vcs_ref))

        self._executil.callExternal([hgBin, 'checkout', '--check',
                                     vcs_ref] + self._hgCheckoutTool(config))

    def vcsCommit(self, config, message, files):
        hgBin = config['env']['hgBin']
        files = files or ['-A']
        self._executil.callExternal(
            [hgBin, 'commit', '-A', '-m', message] + files)

    def vcsTag(self, config, tag, message):
        hgBin = config['env']['hgBin']
        self._executil.callExternal([hgBin, 'tag', '-m', message, tag])

    def vcsPush(self, config, refs):
        refs = refs or []
        env = config['env']
        hgBin = env['hgBin']
        opts = []

        tags = self.vcsListTags(config, '.', '')

        for r in refs:
            if r in tags:
                continue
            if r == '--new-branch':
                opts.append(r)
                continue

            opts.append('-b')
            opts.append(r)
        self._executil.callExternal([hgBin, 'push'] + opts)

    def vcsGetRevision(self, config):
        hgBin = config['env']['hgBin']
        return self._executil.callExternal([hgBin, 'identify', '--id'], verbose=False).strip()

    def _hgCache(self, config, vcs_cache_dir):
        ospath = self._ospath
        hgBin = config['env']['hgBin']
        vcsrepo = config['vcsRepo']

        if vcs_cache_dir is None:
            # Hg does not allow remote repository listing
            if ospath.exists('.hg') and self.vcsGetRepo(config, vcs_cache_dir) == vcsrepo:
                vcs_cache_dir = '.'
            else:
                vcs_cache_dir = self._pathutil.cacheDir('hg')
                vcs_cache_dir = ospath.join(
                    vcs_cache_dir,
                    vcsrepo.replace('/', '_').replace(':', '_')
                )
            vcs_cache_dir = ospath.realpath(vcs_cache_dir)

        if ospath.isdir(vcs_cache_dir):
            remote_info = self.vcsGetRepo(config, vcs_cache_dir)

            if remote_info != vcsrepo:
                self._warn('removing Hg cache on remote URL mismatch: {0} != {1}'
                           .format(remote_info, vcsrepo))
                self._pathutil.rmTree(vcs_cache_dir)
            else:
                self._executil.callExternal(
                    [hgBin, '--cwd', vcs_cache_dir, 'pull'])

        if not ospath.isdir(vcs_cache_dir):
            self._executil.callExternal(
                [hgBin, 'clone', vcsrepo, vcs_cache_dir])

        return vcs_cache_dir

    def vcsGetRefRevision(self, config, vcs_cache_dir, branch):
        vcs_cache_dir = self._hgCache(config, vcs_cache_dir)

        hgBin = config['env']['hgBin']

        res = self._executil.callExternal([
            hgBin, '--repository', vcs_cache_dir, 'branches'
        ], verbose=False).strip().split("\n")

        for r in res:
            r = r.split()

            if r and r[0] == branch:
                return r[1]

        self._errorExit("Uknown Hg ref: {0}".format(branch))

    def vcsListTags(self, config, vcs_cache_dir, tag_hint):
        vcs_cache_dir = self._hgCache(config, vcs_cache_dir)

        hgBin = config['env']['hgBin']

        res = self._executil.callExternal([
            hgBin, '--repository', vcs_cache_dir, 'tags'
        ], verbose=False).strip().split("\n")

        res = [v and v.split()[0] or '' for v in res]
        res = list(filter(None, res))
        del res[res.index('tip')]
        return res

    def vcsListBranches(self, config, vcs_cache_dir, branch_hint):
        vcs_cache_dir = self._hgCache(config, vcs_cache_dir)

        hgBin = config['env']['hgBin']

        res = self._executil.callExternal([
            hgBin, '--repository', vcs_cache_dir, 'branches'
        ], verbose=False).strip().split("\n")

        res = [v and v.split()[0] or '' for v in res]
        res = list(filter(None, res))
        return res

    def vcsExport(self, config, vcs_cache_dir, vcs_ref, dst_path):
        ospath = self._ospath
        os = self._os
        vcs_cache_dir = self._hgCache(config, vcs_cache_dir)

        self._executil.callExternal([
            config['env']['hgBin'],
            '--repository', vcs_cache_dir,
            'archive',
            '--rev', vcs_ref,
            '--type', 'files',
            #'--exclude', '.hg*',
            dst_path
        ])

        for f in ['.hgtags', '.hgignore', '.hg_archival.txt']:
            f = ospath.join(dst_path, f)

            if ospath.exists(f):
                os.remove(f)

    def vcsBranch(self, config, vcs_ref):
        hgBin = config['env']['hgBin']
        self._executil.callExternal([hgBin, 'branch', vcs_ref])
        # We do not use bookmarks, so the only way to actually create the
        # branch.
        self.vcsCommit(config, "CID new branch " + vcs_ref, [])
        self.vcsPush(config, ['--new-branch', vcs_ref])

    def vcsMerge(self, config, vcs_ref, cleanup):
        curr_ref = self._getCurrentBranch(config)

        hgBin = config['env']['hgBin']

        try:
            self._executil.callExternal(
                [hgBin, 'merge', '--tool', 'internal:merge', vcs_ref])
        except self._ext.subprocess.CalledProcessError:
            if cleanup:
                self.vcsRevert(config)
                self._errorExit('Merge failed, aborted.')
            self._errorExit('Merge failed, left as-is.')

        self.vcsCommit(config, "CID merged " + vcs_ref, [])
        self.vcsPush(config, [curr_ref])

    def vcsDelete(self, config, vcs_cache_dir, vcs_ref):
        os = self._os
        vcs_cache_dir = self._hgCache(config, vcs_cache_dir)
        hgBin = config['env']['hgBin']

        old_cwd = os.getcwd()
        os.chdir(vcs_cache_dir)

        self._executil.callExternal([hgBin, 'checkout', '--check',
                                     vcs_ref] + self._hgCheckoutTool(config))
        self.vcsCommit(config, "CID branch delete", ['--close-branch'])
        self.vcsPush(config, [vcs_ref])
        os.chdir(old_cwd)

    def vcsRevert(self, config):
        hgBin = config['env']['hgBin']
        self._executil.callExternal([
            hgBin,
            'update', '--clean',
        ] + self._hgCheckoutTool(config))
        self._executil.callExternal([
            hgBin,
            '--config', 'extensions.purge=',
            'purge', '-I', '*.orig', '-I', '**/*.orig', '--all'
        ])

    def vcsIsMerged(self, config, vcs_ref):
        res = self._executil.callExternal([
            config['env']['hgBin'],
            'merge', '--preview',
            '--tool=internal:merge',
            vcs_ref
        ], verbose=False).strip()
        return res == ''

    def vcsClean(self, config):
        hgBin = config['env']['hgBin']
        self._executil.callExternal([
            hgBin,
            '--config', 'extensions.purge=',
            'purge', '--all',
        ])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext
from . import simple_memo as _simple_memo
from . import complex_memo as _complex_memo
from . import log as _log


@_simple_memo
def isCentOS():
    return _ext.platform.linux_distribution()[0].startswith('CentOS')


@_simple_memo
def isFedora():
    return _ext.ospath.exists('/etc/fedora-release')


@_simple_memo
def isGentoo():
    return _ext.ospath.exists('/etc/gentoo-release')


@_simple_memo
def isArchLinux():
    # There are cases when platform.linux_distribution() is empty on Arch
    return (_ext.ospath.exists('/etc/arch-release') or
            _ext.platform.linux_distribution()[0].startswith('arch'))


@_simple_memo
def isDebian():
    return _ext.platform.linux_distribution()[0].startswith('debian')


@_simple_memo
def isUbuntu():
    return _ext.platform.linux_distribution()[0].startswith('Ubuntu')


@_simple_memo
def isOracleLinux():
    return _ext.platform.linux_distribution()[0].startswith('Oracle Linux')


@_simple_memo
def isRHEL():
    return _ext.platform.linux_distribution()[0].startswith('Red Hat Enterprise Linux')


@_simple_memo
def isOpenSUSE():
    return _ext.platform.linux_distribution()[0].startswith('openSUSE')


@_simple_memo
def isSLES():
    return _ext.platform.linux_distribution()[0].startswith('SUSE Linux Enterprise')


@_simple_memo
def isAlpineLinux():
    return _ext.ospath.exists('/etc/alpine-release')


@_simple_memo
def alpineLinuxVersion():
    return _ext.pathutil.readTextFile('/etc/alpine-release').strip().split('.')


@_simple_memo
def linuxDistVersion():
    if isAlpineLinux():
        return alpineLinuxVersion()

    return _ext.platform.linux_distribution()[1]


def linuxDistMajorVer():
    return linuxDistVersion().split('.')[0]


@_simple_memo
def isLinux():
    return _ext.platform.system() == 'Linux'


@_simple_memo
def isMacOS():
    return _ext.platform.system() == 'Darwin'


@_simple_memo
def isAMD64():
    return _ext.platform.machine() == 'x86_64'


@_simple_memo
def basearch():
    return _ext.platform.machine()


@_simple_memo
def osCodeName():
    codename = _ext.subprocess.check_output(['lsb_release', '-cs'])
    codename = _ext.executil.toString(codename)
    return codename.strip()


@_simple_memo
def osRelease():
    codename = _ext.subprocess.check_output(['lsb_release', '-rs'])
    codename = _ext.executil.toString(codename)
    return codename.strip()


def isAdmin():
    return _ext.os.geteuid() == 0


@_complex_memo
def haveGroup(grpname):
    try:
        gid = _ext.grp.getgrnam(grpname)[2]
    except KeyError as e:
        _log.warn(str(e))
        return False

    return gid in _ext.os.getgroups()


@_simple_memo
def isSCLSupported():
    "Check if Software Collections are supported"
    return (
        isCentOS() or
        isRHEL() or
        isOracleLinux()
    )


@_simple_memo
def installerType():
    # Known OS
    #---
    if isDebian() or isUbuntu():
        return 'apt'

    if isArchLinux():
        return 'pacman'

    if isFedora():
        return 'dnf'

    if isCentOS() or isRHEL() or isOracleLinux():
        return 'yum'

    if isAlpineLinux():
        return 'apk'

    if isGentoo():
        return 'emerge'

    if isOpenSUSE() or isSLES():
        return 'zypper'

    if isMacOS():
        return 'brew'

    # Fallback
    #---
    pathutil = _ext.pathutil

    if pathutil.which('apt-get'):
        return 'apt'

    if pathutil.which('dnf'):
        return 'dnf'

    if pathutil.which('zypper'):
        return 'zypper'

    if pathutil.which('yum'):
        return 'yum'

    if pathutil.which('pacman'):
        return 'pacman'

    if _ext.ospath.exists('/sbin/apk'):
        return 'apk'

    if pathutil.which('emerge'):
        return 'emerge'

    return None


@_simple_memo
def packageType():
    installer = installerType()

    if installer == 'apt':
        return 'deb'

    if installer in ('dnf', 'zypper', 'yum'):
        return 'rpm'

    if installer == 'emerge':
        return 'ebuild'

    return installer


def isDeb():
    return packageType() == 'deb'


def isRPM():
    return packageType() == 'rpm'


def isEmerge():
    return installerType() == 'emerge'


def isApk():
    return installerType() == 'apk'


def isPacman():
    return installerType() == 'pacman'


def isBrew():
    return installerType() == 'brew'


def isYum():
    return installerType() == 'yum'


def isDnf():
    return installerType() == 'dnf'


def isZypper():
    return installerType() == 'zypper'


def isDisabledToolsSetup(env):
    val = env['externalSetup']
    return (
        val == True or
        val == '1' or
        val == 'yes' or
        val == 'on'
    )


def isExternalToolsSetup(env):
    val = env['externalSetup']
    return val != False and not isDisabledToolsSetup(env)


def autoDetectByCfg(name, config, file_name):
    if name in config.get('toolOrder', []):
        return True

    if type(file_name) is type(''):
        file_name = [file_name]

    root_list = config['projectRootSet']

    for f in file_name:
        if f in root_list:
            return True

    return False



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .subtool import SubTool

__all__ = ['VcsTool']


class VcsTool(SubTool):
    __slots__ = ()

    def autoDetect(self, config):
        return self._autoDetectVCS(config, self.autoDetectFiles())

    def vcsGetRepo(self, config, wc_dir=None):
        raise NotImplementedError(self._name)

    def vcsCheckout(self, config, branch):
        raise NotImplementedError(self._name)

    def vcsCommit(self, config, message, files):
        raise NotImplementedError(self._name)

    def vcsTag(self, config, tag, message):
        raise NotImplementedError(self._name)

    def vcsPush(self, config, refs):
        raise NotImplementedError(self._name)

    def vcsGetRevision(self, config):
        raise NotImplementedError(self._name)

    def vcsGetRefRevision(self, config, vcs_cache_dir, branch):
        raise NotImplementedError(self._name)

    def vcsListTags(self, config, vcs_cache_dir, tag_hint):
        raise NotImplementedError(self._name)

    def vcsListBranches(self, config, vcs_cache_dir, branch_hint):
        raise NotImplementedError(self._name)

    def vcsExport(self, config, vcs_cache_dir, vcs_ref, dst_path):
        raise NotImplementedError(self._name)

    def vcsBranch(self, config, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsMerge(self, config, vcs_ref, cleanup):
        raise NotImplementedError(self._name)

    def vcsDelete(self, config, vcs_cache_dir, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsRevert(self, config):
        raise NotImplementedError(self._name)

    def vcsIsMerged(self, config, vcs_ref):
        raise NotImplementedError(self._name)

    def vcsClean(self, config):
        raise NotImplementedError(self._name)

    def _autoDetectVCS(self, config, vcsDir):
        if config.get('vcs', None) == self._name:
            return True

        if vcsDir in config['projectRootSet']:
            if config.get('vcs', self._name) != self._name:
                self._errorExit(
                    'Another VCS type {0} has been already detected!'.format(config['vcs']))
            config['vcs'] = self._name
            return True

        return False



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .mixins.ondemand import OnDemandMixIn
from .mixins.log import LogMixIn

__all__ = ['SubTool']


class SubTool(LogMixIn, OnDemandMixIn):
    __slots__ = (
        '_name',
        '_have_tool',
    )

    # TODO: get rid of
    SYSTEM_VER = 'system'

    def __init__(self, name):
        super(SubTool, self).__init__()
        self._name = name
        self._have_tool = False

    def getDeps(self):
        return []

    def getVersionParts(self):
        """Override, if tool has different number of meaningful parts"""
        return 2

    def envDeps(self, env):
        pass

    def getPostDeps(self):
        return []

    def getOrder(self):
        return 0

    def _installTool(self, env):
        self._errorExit(
            'Tool "{0}" must be manually installed'.format(self._name))

    def envNames(self):
        return [self._name + 'Bin']

    def importEnv(self, env):
        environ = self._environ

        for name in self.envNames():
            val = environ.get(name, None)
            if val is not None:
                env[name] = val

    def exportEnv(self, env, dst):
        for name in self.envNames():
            if name in env:
                dst[name] = env[name]

    def initEnv(self, env, bin_name=None):
        name = self._name
        bin_env = name + 'Bin'

        if not env.get(bin_env, None):
            if bin_name is None:
                bin_name = name

            tool_path = self._pathutil.which(bin_name)
            if tool_path:
                env[bin_env] = tool_path.strip()
                self._have_tool = True
        else:
            self._have_tool = True

    def autoDetect(self, config):
        files = self.autoDetectFiles()

        if files:
            return self._detect.autoDetectByCfg(self._name, config, files)

        return False

    def autoDetectFiles(self):
        return None

    def requireInstalled(self, env):
        self.importEnv(env)
        self.initEnv(env)
        self.installTool(env)

    def installTool(self, env):
        if not self._have_tool:
            if self._detect.isDisabledToolsSetup(env):
                self._errorExit(
                    'Tool "{0}" must be installed externally (env config)'.format(self._name))
            else:
                self._warn(
                    'Auto-installing required "{0}" tool'.format(self._name))

                if self._detect.isExternalToolsSetup(env):
                    self._executil.externalSetup(env, [
                        'tool', 'install',
                        self._name,
                        env.get(self._name + 'Ver', '')])
                    self._afterExternalSetup(env)
                else:
                    self._installTool(env)

            self.initEnv(env)

            if not self._have_tool:
                self._errorExit('Failed to install "{0}"'.format(self._name))

    def _afterExternalSetup(self, env):
        pass

    def isInstalled(self, env):
        self.initEnv(env)
        return self._have_tool

    def updateTool(self, env):
        if self._detect.isDisabledToolsSetup(env):
            self._errorExit(
                'Tool "{0}" must be updated externally (env config)'.format(self._name))
        elif self._detect.isExternalToolsSetup(env):
            self._executil.externalSetup(env, [
                'tool', 'update',
                self._name,
                env.get(self._name + 'Ver', '')])
        else:
            self._updateTool(env)

    def _updateTool(self, env):
        self.requireInstalled(env)

    def uninstallTool(self, env):
        self._have_tool = False
        self._errorExit(
            'Tool "{0}" must be uninstalled externally'.format(self._name))

    def loadConfig(self, config):
        pass

    def updateProjectConfig(self, config, updates):
        """
updates = {
    name : '...',
    version : '...',
}
@return a list of files to be committed
"""
        return []

    def onExec(self, env, args, replace=True):
        bin = env.get(self._name + 'Bin', None)

        if bin:
            self._executil.callInteractive([bin] + args, replace=replace)
        else:
            self._errorExit(
                'Exec command has not been implemented for "{0}"'.format(self._name))

    def _getTune(self, config, key, default=None):
        return config.get('toolTune', {}).get(self._name, {}).get(key, default)

    def sanitizeVersion(self, env):
        """Should be called implicitly by standard CID functionality"""
        ver_var = self._name + 'Ver'
        ver = env.get(ver_var, None)

        if ver:
            end = self.getVersionParts()
            new_ver = ver.split('.')[:end]
            new_ver = '.'.join(new_ver)

            if env[ver_var] != new_ver:
                self._warn('Too precise version "{0}" for "{1}" - trimmed to "{2}"'
                           .format(ver, self._name, new_ver))
                env[ver_var] = new_ver

    def onConfigReset(self):
        self._have_tool = False



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..migrationtool import MigrationTool


class flywayTool(MigrationTool):
    """Flyway - Evolve your Database Schema easily and
reliably across all your instances.

Home: https://flywaydb.org/

Migration is run automatically, if flyway.conf is found.

FLYWAY_HOME is user/set similar to LIQUIBASE_HOME

Note: for now, it gets the latest tag from GitHub and then uses
    maven repo for download. That obviously should be improved.
"""
    __slots__ = ()

    FLYWAY_CONF = 'flyway.conf'

    def autoDetectFiles(self):
        return self.FLYWAY_CONF

    def getDeps(self):
        return ['java', 'tar', 'gzip']

    def getVersionParts(self):
        return 3

    def envNames(self):
        return ['flywayVer', 'flywayDir', 'flywayDrivers']

    def _installTool(self, env):
        ospath = self._ospath
        os = self._os
        github = self._ext.github

        lb_dir = env['flywayDir']
        ver = env.get('flywayVer', 'latest')

        if ver == 'latest':
            release = ver
        else:
            release = 'tags/flyway-{0}'.format(ver)

        info = github.releaseInfo(env, 'flyway/flyway', release)

        found_ver = info['tag_name'].replace('flyway-', '')
        dst = ospath.join(lb_dir, found_ver)

        if not ospath.exists(dst):
            maven_central = self._install.mavenCentral(env)
            asset_url = '{0}/org/flywaydb/flyway-commandline/{1}/flyway-commandline-{1}.tar.gz'.format(
                maven_central, found_ver
            )
            self._pathutil.downloadExtract(
                env, asset_url,
                dst, 'z', 1)

        if ver == 'latest':
            latest = ospath.join(lb_dir, 'latest')
            try:
                os.unlink(latest)
            except OSError:
                pass
            os.symlink(found_ver, latest)

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        ver = env.get('flywayVer', 'latest')
        inst_dir = env['flywayDir']

        if ver != 'latest':
            inst_dir = self._pathutil.safeJoin(inst_dir, ver)

        self._pathutil.rmTree(inst_dir)

    def initEnv(self, env):
        ospath = self._ospath

        #---
        inst_dir = self._environ.get('FLYWAY_HOME', None)
        fail_on_missing = False

        if inst_dir:
            fail_on_missing = True
        else:
            lb_dir = ospath.join(
                self._pathutil.userHome(),
                '.local', 'flywaybin')
            lb_dir = env.setdefault('flywayDir', lb_dir)
            ver = env.get('flywayVer', 'latest')
            inst_dir = ospath.join(lb_dir, ver)

        lb_bin = ospath.join(inst_dir, 'flyway')

        if ospath.exists(lb_bin):
            env['flywayBin'] = lb_bin
            self._environ['FLYWAY_HOME'] = inst_dir
            self._have_tool = True
        elif fail_on_missing:
            self._errorExit('Unset FLYWAY_HOME or '
                            'setup the tool manually.')

    def onMigrate(self, config):
        if self._ospath.exists(self.FLYWAY_CONF):
            self.onExec(config['env'], ['-n', 'migrate'], replace=False)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..runenvtool import RunEnvTool
from ..runtimetool import RuntimeTool
from .curltoolmixin import CurlToolMixIn


class dockerTool(BuildTool, CurlToolMixIn, RunEnvTool, RuntimeTool):
    """Docker - Build, Ship, and Run Any App, Anywhere.

Home: https://www.docker.com/

Experimental support.

Docker CE support is added for CentOS, Fedora, Debian and Ubuntu.
For other systems, "docker" or "docker-engine" packages is tried to be installed.

Docker EE or other installation methods are out of scope for now.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'Dockerfile'

    def getOrder(self):
        return 10

    def envNames(self):
        return ['dockerBin', 'dockerVer', 'dockerRepos', 'dockerTag']

    def _installTool(self, env):
        detect = self._detect
        install = self._install
        repo = env.get('dockerRepos', 'https://download.docker.com')

        if detect.isCentOS():
            install.yumRepo(
                'docker', repo + '/linux/centos/docker-ce.repo')

        elif detect.isFedora():
            install.yumRepo(
                'docker', repo + '/linux/fedora/docker-ce.repo')

        elif detect.isDebian():
            gpg = self._callCurl(env, [repo + '/linux/debian/gpg'])
            install.aptRepo(
                'docker',
                'deb [arch=amd64] {0}/linux/debian $codename$ stable'.format(
                    repo),
                gpg
            )

        elif detect.isUbuntu():
            gpg = self._callCurl(env, [repo + '/linux/ubuntu/gpg'])
            install.aptRepo(
                'docker',
                'deb [arch=amd64] {0}/linux/ubuntu $codename$ stable'.format(
                    repo),
                gpg,
                codename_map={},
                repo_base=repo,
            )

        elif detect.isRHEL():
            install.yumRHELRepo('server-extras-rpms')
            install.yum(['docker'])
            self._executil.startService('docker')
            return

        elif detect.isOracleLinux():
            install.yumOLPublic('addons')
            install.yum(['docker'])
            self._executil.startService('docker')
            return

        elif detect.isSLES():
            install.SUSEConnectVerArch('sle-module-containers')
            install.zypper('docker')
            self._executil.startService('docker')
            return

        else:
            install.yumEPEL()
            install.debrpm(['docker'])
            install.debrpm(['docker-engine'])
            install.emerge(['app-emulation/docker'])
            install.pacman(['docker'])
            install.apkCommunity()
            install.apk(['docker'])

            self._executil.startService('docker')

            return

        ver = env.get('dockerVer', None)

        if ver:
            install.debrpm(['docker-ce-' + ver])
        else:
            install.debrpm(['docker-ce'])

        self._executil.startService('docker')

    def onBuild(self, config):
        ospath = self._ospath

        env = config['env']
        tag = env.get('dockerTag', ospath.basename(ospath.realpath('.')))
        cmd = [env['dockerBin'], 'build', '-t', tag, '.']

        if not self._isDockerAdmin():
            sudo = self._pathutil.which('sudo')
            cmd = [sudo] + cmd

        self._executil.callMeaningful(cmd)

    def onExec(self, env, args, replace=True):
        bin = env['dockerBin']

        if self._isDockerAdmin():
            cmd = [bin] + args
        else:
            sudo = self._pathutil.which('sudo')
            cmd = [sudo, bin] + args

        self._executil.callInteractive(cmd, replace=replace)

    def onRun(self, config, svc, args):
        env = config['env']
        cmd = [env['dockerBin'], 'run', svc['path']]

        if not self._isDockerAdmin():
            sudo = self._pathutil.which('sudo')
            cmd = [sudo] + cmd

        self._executil.callMeaningful(cmd)

    def _isDockerAdmin(self):
        detect = self._detect
        return detect.haveGroup('docker') or detect.isAdmin()

    def tuneDefaults(self, env):
        return {
            'minMemory': '256M',
            'debugOverhead': '128M',
            'connMemory': '100K',
            'debugConnOverhead': '1M',
            'socketType': 'tcp',
            'scalable': False,
            'reloadable': False,
            'multiCore': True,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn
from .curltoolmixin import CurlToolMixIn


class gvmTool(BashToolMixIn, CurlToolMixIn, RunEnvTool):
    """Go Version Manager.

Home: https://github.com/moovweb/gvm
"""
    __slots__ = ()

    GVM_VERSION_DEFAULT = 'master'
    GVM_INSTALLER_DEFAULT = 'https://raw.githubusercontent.com/moovweb/gvm/master/binscripts/gvm-installer'

    def getDeps(self):
        return (
            ['git', 'hg', 'make', 'binutils'] +
            BashToolMixIn.getDeps(self) +
            CurlToolMixIn.getDeps(self))

    def _installTool(self, env):
        self._install.deb(['bison', 'gcc', 'build-essential'])
        self._install.rpm(['bison', 'gcc', 'glibc-devel'])
        self._install.emergeDepsOnly(['dev-lang/go'])
        self._install.pacman(['bison', 'gcc', 'glibc', ])
        self._install.apk('bison')
        self._builddep.essential()

        gvm_installer = self._callCurl(env, [env['gvmInstaller']])
        self._callBash(
            env,
            input=gvm_installer,
            suppress_fail=True)  # error when Go is not yet installed

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        gvm_dir = env['gvmDir']
        self._pathutil.rmTree(gvm_dir)
        self._have_tool = False

    def envNames(self):
        return ['gvmDir', 'gvmInstaller']

    def initEnv(self, env):
        ospath = self._ospath
        os = self._os
        environ = self._environ
        gvm_dir = ospath.join(environ['HOME'], '.gvm')
        gvm_dir = env.setdefault('gvmDir', gvm_dir)
        environ['GVM_DEST'] = ospath.dirname(gvm_dir)
        environ['GVM_NAME'] = ospath.basename(gvm_dir)
        environ['GVM_NO_UPDATE_PROFILE'] = '1'

        env.setdefault('gvmVer', self.GVM_VERSION_DEFAULT)
        env.setdefault('gvmInstaller', self.GVM_INSTALLER_DEFAULT)

        env_init = ospath.join(gvm_dir, 'scripts', 'gvm')
        env['gvmInit'] = env_init

        self._have_tool = ospath.exists(env_init)

    def onExec(self, env, args, replace=True):
        cmd = '. {0} && gvm {1}'.format(
            env['gvmInit'], self._ext.subprocess.list2cmdline(args))
        self._callBashInteractive(env, cmd, replace=replace)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class pipTool(BuildTool):
    """The PyPA recommended tool for installing Python packages.

Home: https://pypi.python.org/pypi/pip
"""
    __slots__ = ()

    REQUIREMENTS_FILES = [
        'requirements.txt',
        'requirements.pip',
    ]

    def autoDetectFiles(self):
        return self.REQUIREMENTS_FILES

    def getDeps(self):
        return ['python', 'virtualenv']

    def envNames(self):
        return ['pipBin', 'pipVer']

    def _installTool(self, env):
        ospath = self._ospath

        if ospath.exists(env['pipBin']):
            self._updateTool(env)
        else:
            self._executil.callExternal([
                ospath.join(env['virtualenvDir'], 'bin',
                            'easy_install'), 'pip'
            ])

    def _updateTool(self, env):
        self._executil.callExternal([
            env['pipBin'], 'install', '-q',
            '--upgrade',
            'pip>={0}'.format(env['pipVer']),
        ])

    def installTool(self, env):
        if not self._have_tool:
            self._installTool(env)
            self.initEnv(env)

            if not self._have_tool:
                self._errorExit('Failed to install "{0}"'.format(self._name))

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        pipBin = ospath.join(env['virtualenvDir'], 'bin', 'pip')
        pipBin = env.setdefault('pipBin', pipBin)
        pipVer = env.setdefault('pipVer', '9.0.3')

        if ospath.exists(pipBin):
            pipFactVer = self._executil.callExternal(
                [pipBin, '--version'], verbose=False)
            pipFactVer = [int(v) for v in pipFactVer.split(' ')[1].split('.')]
            pipNeedVer = [int(v) for v in pipVer.split('.')]

            self._have_tool = pipNeedVer <= pipFactVer

    def onPrepare(self, config):
        for rf in self.REQUIREMENTS_FILES:
            if self._ospath.exists(rf):
                cmd = [config['env']['pipBin'], 'install', '-r', rf]
                self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..subtool import SubTool


class NpmToolMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['node', 'npm']

    def _npmName(self):
        return self._name

    def _isGlobalNpm(self):
        return self._detect.isAlpineLinux()

    def _installTool(self, env):
        cmd = [env['npmBin'], 'install', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd + ['--unsafe-perm'])
        else:
            self._executil.callExternal(cmd)

    def _updateTool(self, env):
        cmd = [env['npmBin'], 'update', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd + ['--unsafe-perm'])
        else:
            self._executil.callExternal(cmd)

    def uninstallTool(self, env):
        cmd = [env['npmBin'], 'uninstall', '-g', self._npmName()]

        if self._isGlobalNpm():
            self._executil.trySudoCall(cmd + ['--unsafe-perm'])
        else:
            self._executil.callExternal(cmd)

        self._have_tool = False

    def initEnv(self, env, bin_name=None):
        name = self._name
        bin_env = name + 'Bin'

        if not env.get(bin_env, None):
            if bin_name is None:
                bin_name = name

            npmBinDir = env.get('npmBinDir', None)

            if npmBinDir:
                tool_path = self._pathutil.safeJoin(npmBinDir, bin_name)

                if self._ospath.exists(tool_path):
                    env[bin_env] = tool_path
                    self._have_tool = True
        else:
            self._have_tool = True



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class gzipTool(BuildTool):
    """Compression utility designed to be a replacement for compress.

Home: http://www.gzip.org/

Auto-detected, if webcfg has mount points serving static files with gzip
static enabled.

Tune:
* toGzipRe = '\.(js|json|css|svg|txt|xml|html)$'
"""
    __slots__ = ()
    __TO_GZIP = '\.(js|json|css|svg|txt|xml|html|md)$'

    def autoDetect(self, config):
        if not config.get('packageGzipStatic', True):
            return False

        webcfg = config.get('webcfg', {})
        webroot = webcfg.get('root', '.')

        for (m, v) in webcfg.get('mounts', {}).items():
            if not isinstance(v, dict):
                continue

            if v.get('static', False):
                pass
            elif v.get('app', None) is None:
                pass
            else:
                continue

            if v.get('tune', {}).get('staticGzip', True):
                return True

        return False

    def _installTool(self, env):
        self._install.debrpm(['gzip'])
        self._install.emerge(['app-arch/gzip'])
        self._install.pacman(['gzip'])
        self._install.apk(['gzip'])
        self._install.brew('gzip')

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/bin/gzip'):
            return

        super(gzipTool, self).initEnv(env, bin_name)

    def onBuild(self, config):
        if not self.autoDetect(config):
            return

        self._info('Generating GZip files of static content')
        webcfg = config.get('webcfg', {})
        webroot = webcfg.get('root', '.')
        re = self._ext.re
        to_gzip_re = self._getTune(config, 'toGzipRe', self.__TO_GZIP)
        to_gzip_re = re.compile(to_gzip_re, re.I)

        ospath = self._ospath
        gzip = self._ext.gzip
        shutil = self._ext.shutil

        for (m, v) in webcfg.get('mounts', {}).items():
            if not isinstance(v, dict):
                continue

            if v.get('static', False):
                pass
            elif v.get('app', None) is None:
                pass
            else:
                continue

            if not v.get('tune', {}).get('staticGzip', True):
                continue

            gzip_dir = ospath.join(webroot, m)
            gzip_dir = gzip_dir.lstrip('/')
            self._info('> ' + gzip_dir)

            for (path, dirs, files) in self._os.walk(gzip_dir):
                for f in files:
                    if to_gzip_re.search(f):
                        f = ospath.join(path, f)
                        with open(f, 'rb') as f_in:
                            with gzip.open(f + '.gz', 'wb', 9) as f_out:
                                shutil.copyfileobj(f_in, f_out)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from ..rmstool import RmsTool


class scpTool(RmsTool):
    """secure copy.

RMS repo formats:

ssh://user@host/port:root/path - full remote, note the port
user@host:root/path - short remote
user@host - shortest remote
/some/path - local path
../some/other/path - another local path

More details:
* Packages are set read-only
* SSH shell access is required to remote as well (restricted shell is not suitable)
* SCP itself is used only for upload & retrieval. The rest is done through SSH.
* {hash_type}sum utilities must be present on remote for hash calculation support.
"""
    __slots__ = ()

    REMOTE_PATTERN = '^(ssh://)?([a-zA-Z][a-zA-Z0-9_]+@[a-zA-Z0-9_.]+)(/([0-9]{1,5}))(:(.+))?$'
    REMOTE_GRP_USER_HOST = 2
    REMOTE_GRP_PORT = 4
    REMOTE_GRP_PATH = 6

    def getDeps(self):
        return ['ssh']

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath
        re = self._ext.re
        scpBin = config['env']['scpBin']
        rms_repo = config['rmsRepo']

        remote = re.match(self.REMOTE_PATTERN, rms_repo)
        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = remote.group(self.REMOTE_GRP_PATH)

        for package in package_list:
            package_basename = ospath.basename(package)

            if remote:
                dst = "{0}:{1}".format(
                    user_host,
                    ospath.join(path, rms_pool, package_basename)
                )

                if '/' in rms_pool:
                    cmd = "mkdir -p {0}".format(ospath.join(path, rms_pool))
                    self._callSSH(config, user_host, port, cmd)

                self._callRemoteSCP(config, port, package, dst)

                cmd = "chmod ugo-wx {0}".format(
                    ospath.join(path, rms_pool, package_basename))
                self._callSSH(config, user_host, port, cmd)
            else:
                dst = ospath.join(rms_repo, rms_pool, package_basename)

                if '/' in rms_pool:
                    self._executil.callExternal(
                        ['mkdir', '-p', ospath.join(rms_repo, rms_pool)])

                self._executil.callExternal([scpBin, '-Bq', package, dst])
                self._executil.callExternal(['chmod', 'ugo-wx', dst])

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        ospath = self._ospath
        re = self._ext.re
        scpBin = config['env']['scpBin']
        rms_repo = config['rmsRepo']

        remote = re.match(self.REMOTE_PATTERN, rms_repo)
        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = remote.group(self.REMOTE_GRP_PATH)

        for package in package_list:
            package_basename = ospath.basename(package)

            if remote:
                if '/' in dst_pool:
                    cmd = "mkdir -p {0}".format(ospath.join(path, dst_pool))
                    self._callSSH(config, user_host, port, cmd)

                cmd = 'cp -a {0} {1} && chmod ugo-wx {1}'.format(
                    ospath.join(path, src_pool, package_basename),
                    ospath.join(path, dst_pool, package_basename)
                )
                self._callSSH(config, user_host, port, cmd)
            else:
                src = ospath.join(rms_repo, src_pool, package_basename)
                dst = ospath.join(rms_repo, dst_pool, package_basename)

                if '/' in dst_pool:
                    self._executil.callExternal(
                        ['mkdir', '-p', ospath.join(rms_repo, dst_pool)])

                self._executil.callExternal([scpBin, '-Bq', src, dst])
                self._executil.callExternal(['chmod', 'ugo-wx', dst])

    def rmsGetList(self, config, rms_pool, package_hint):
        ospath = self._ospath
        re = self._ext.re

        env = config['env']
        rms_repo = config['rmsRepo']
        remote = re.match(self.REMOTE_PATTERN, rms_repo)

        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = ospath.join(remote.group(self.REMOTE_GRP_PATH), rms_pool)
            cmd = 'ls {0}'.format(path)
            ret = self._callSSH(config, user_host, port, cmd).split("\n")
        else:
            path = ospath.join(rms_repo, rms_pool)
            ret = self._os.listdir(path)

        return ret

    def rmsRetrieve(self, config, rms_pool, package_list):
        ospath = self._ospath
        re = self._ext.re

        scpBin = config['env']['scpBin']
        rms_repo = config['rmsRepo']

        remote = re.match(self.REMOTE_PATTERN, rms_repo)
        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = remote.group(self.REMOTE_GRP_PATH)

        for package in package_list:
            package_basename = ospath.basename(package)

            if remote:
                src = "{0}:{1}".format(
                    user_host,
                    ospath.join(path, rms_pool, package_basename)
                )
                self._callRemoteSCP(config, port, src, package_basename)
            else:
                src = ospath.join(rms_repo, rms_pool, package_basename)
                self._executil.callExternal(
                    [scpBin, '-Bq', src, package_basename])

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        ospath = self._ospath
        re = self._ext.re

        env = config['env']
        rms_repo = config['rmsRepo']
        remote = re.match(self.REMOTE_PATTERN, rms_repo)

        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = ospath.join(remote.group(
                self.REMOTE_GRP_PATH), rms_pool, package)
            cmd = "{0}sum {1}".format(hash_type, path)
            ret = self._callSSH(config, user_host, port, cmd,
                                verbose=False).strip().split()[0]
        else:
            path = ospath.join(rms_repo, rms_pool, package)
            ret = self.rmsCalcHash(path, hash_type).split(':', 1)[1]

        return ret

    def rmsPoolCreate(self, config, rms_pool):
        ospath = self._ospath
        re = self._ext.re

        env = config['env']
        rms_repo = config['rmsRepo']
        remote = re.match(self.REMOTE_PATTERN, rms_repo)

        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = ospath.join(remote.group(self.REMOTE_GRP_PATH), rms_pool)
            cmd = 'mkdir -p {0}'.format(path)
            self._callSSH(config, user_host, port, cmd)
        else:
            path = ospath.join(rms_repo, rms_pool)

            if not ospath.exists(path):
                self._os.makedirs(path)

    def rmsPoolList(self, config):
        os = self._os
        re = self._ext.re

        env = config['env']
        rms_repo = config['rmsRepo']
        remote = re.match(self.REMOTE_PATTERN, rms_repo)

        if remote:
            user_host = remote.group(self.REMOTE_GRP_USER_HOST)
            port = remote.group(self.REMOTE_GRP_PORT)
            path = remote.group(self.REMOTE_GRP_PATH)
            cmd = 'ls {0}'.format(path)
            ret = self._callSSH(config, user_host, port, cmd).split("\n")
        else:
            ret = os.listdir(rms_repo)

        return ret

    def _callRemoteSCP(self, config, port, src, dst):
        port = port or '22'
        env = config['env']
        self._executil.callExternal([
            env['scpBin'],
            '-Bq', '-P', port,
            '-o', 'StrictHostKeyChecking={0}'.format(
                env['sshStrictHostKeyChecking']),
            src, dst
        ])

    def _callSSH(self, config, user_host, port, cmd, **kwargs):
        port = port or '22'
        env = config['env']
        return self._executil.callExternal([
            env['sshBin'],
            '-Tqn',
            '-o', 'BatchMode=yes',
            '-o', 'StrictHostKeyChecking={0}'.format(
                env['sshStrictHostKeyChecking']),
            '-p', port,
            user_host, '--', cmd
        ], **kwargs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool


class tarTool(RunEnvTool):
    """The GNU version of the tar archiving utility.
"""
    __slots__ = ()

    def _installTool(self, env):
        self._install.debrpm(['tar'])
        self._install.emerge(['app-arch/tar'])
        self._install.pacman(['tar'])
        self._install.apk(['tar'])
        self._install.brew('gnu-tar')

    def initEnv(self, env, bin_name=None):
        # Busybox's version is not enough for SDKMan
        if self._detect.isAlpineLinux() and self._ospath.islink('/bin/tar'):
            return

        super(tarTool, self).initEnv(env, bin_name)



#
# Copyright 2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..testtool import TestTool


class ctestTool(TestTool):
    """Build, Test and Package Your Software With CMake.

Home: https://cmake.org/

CTest runs tests in CMake build folder, if any.
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'CMakeLists.txt'

    def getDeps(self):
        return ['cmake']

    def envNames(self):
        return ['ctestBin']

    def initEnv(self, env):
        self._environ.setdefault('CTEST_OUTPUT_ON_FAILURE', '1')
        super(ctestTool, self).initEnv(env)

    def onCheck(self, config):
        ospath = self._ospath
        os = self._os
        build_dir = config['env']['cmakeBuildDir']

        if ospath.exists(build_dir):
            os.chdir(build_dir)
            cmd = [config['env']['ctestBin']]
            self._executil.callMeaningful(cmd)
            os.chdir(config['wcDir'])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from ..rmstool import RmsTool


class npmTool(BuildTool, RmsTool):
    """npm is the package manager for JavaScript.

Home: https://www.npmjs.com/

In RMS mode support tuning through .toolTune.npm:
    .tag - npm publish --tag
    .access - npm publish --access

Note: it auto-disables, if Yarn tool is detected
"""
    __slots__ = ()

    PACKAGE_JSON = 'package.json'

    def autoDetectFiles(self):
        return self.PACKAGE_JSON

    def getDeps(self):
        return ['node']

    def _isGlobalNpm(self):
        return self._detect.isAlpineLinux()

    def _isYarnInUse(self, config):
        return 'yarn' in config['toolOrder']

    def _installTool(self, env):
        if self._isGlobalNpm():
            if env['nodeVer'] == 'current':
                self._install.apk('nodejs-npm-current')
            else:
                self._install.apk('nodejs-npm')
            return

    def _updateTool(self, env):
        if self._isGlobalNpm():
            return

        self._executil.callExternal([env['npmBin'], 'update', '-g', 'npm'])

    def uninstallTool(self, env):
        pass

    def initEnv(self, env, bin_name=None):
        # prevent extra background processes
        self._environ['NO_UPDATE_NOTIFIER'] = '1'

        super(npmTool, self).initEnv(env, bin_name)

        if self._have_tool:
            env['npmBinDir'] = self._executil.callExternal(
                [env['npmBin'], 'bin', '-g'], verbose=False).strip()

    def loadConfig(self, config):
        content = self._pathutil.loadJSONConfig(self.PACKAGE_JSON)
        if content is None:
            return

        for f in ('name', 'version'):
            if f in content and f not in config:
                config[f] = content[f]

    def updateProjectConfig(self, config, updates):
        def updater(json):
            for f in ('name', 'version'):
                if f in updates:
                    json[f] = updates[f]

        return self._pathutil.updateJSONConfig(self.PACKAGE_JSON, updater)

    def onPrepare(self, config):
        if self._ospath.exists(self.PACKAGE_JSON) and not self._isYarnInUse(config):
            npmBin = config['env']['npmBin']
            cmd = [npmBin, 'install', '--production=false']
            self._executil.callMeaningful(cmd)

    def _npmPackageName(self, config):
        name = config['name'].replace('@', '').replace('/', '-')

        package = '{0}-{1}.tgz'.format(name, config['version'])

        return package

    def onPackage(self, config):
        npmBin = config['env']['npmBin']

        if self._ospath.exists(self.PACKAGE_JSON) and not self._isYarnInUse(config):
            #self._executil.callExternal([npmBin, 'prune', '--production'])
            # https://github.com/npm/npm/issues/17781
            self._pathutil.rmTree('node_modules')
            cmd = [npmBin, 'install', '--production']
            self._executil.callMeaningful(cmd)

        if config.get('rms', None) == self._name:
            cmd = [npmBin, 'pack']
            self._executil.callMeaningful(cmd)
            package = self._npmPackageName(config)
            self._pathutil.addPackageFiles(config, package)

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath
        package = self._npmPackageName(config)

        if (len(package_list) == 1 and
                ospath.basename(package_list[0]) == package):
            self._warn('Workaround for known "npm publish" tarball issues')
            self._pathutil.rmTree(package)
        else:
            self._errorExit('Unexpected package list: {0}'
                            .format(package_list))

        npmBin = config['env']['npmBin']
        cmd = [npmBin, 'publish']

        tune = config.get('toolTune', {}).get('npm', {})

        if 'tag' in tune:
            cmd += ['--tag', tune['tag']]

        if 'access' in tune:
            cmd += ['--access', tune['access']]

        self._executil.callMeaningful(cmd)

    def rmsPoolList(self, config):
        return [
            'registry',
        ]



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..rmstool import RmsTool
from .piptoolmixin import PipToolMixIn


class twineTool(PipToolMixIn, RmsTool):
    """Collection of utilities for interacting with PyPI
Home: https://pypi.python.org/pypi/twine

Not a fully supported RMS. Supports only uploading of local packages.

Note: rmsRepo is ignored and rmsPool is actual repo URL or uses ~/.pypirc entry
"""
    __slots__ = ()

    def getDeps(self):
        return super(twineTool, self).getDeps() + ['gpg']

    def envNames(self):
        return [
            'twineBin',
            'twineIdentity',
            'twineUser',
            'twinePassword',
        ]

    def rmsUpload(self, config, rms_pool, package_list):
        env = config['env']

        identity = env.get('twineIdentity', None)
        user = env.get('twineUser', None)
        passwd = env.get('twinePassword', None)

        cmd = [env['twineBin'], 'upload']
        cmd += ['--repository', rms_pool]

        if identity:
            cmd += ['--sign', '--identity', identity]
        elif self._getTune(config, 'requireSign', True):
            self._errorExit(
                'Twine config requires GPG signature. Please set "twineIdentity"')

        if user:
            cmd += ['--username', user]

        if passwd:
            cmd += ['--password', passwd]

        cmd.append('--skip-existing')

        cmd += package_list

        self._executil.callExternal(cmd, user_interaction=True)

    def rmsPoolList(self, config):
        return [
            'pypi',
            'pypitest',
        ]



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .data import DataSlots

from collections import OrderedDict


class ConfigMixIn(DataSlots):
    try:
        __str_type = (str, unicode)
    except NameError:
        __str_type = (str, )

    CONFIG_VARS = OrderedDict([
        ('name', __str_type),
        ('version', __str_type),
        ('vcs', __str_type),
        ('vcsRepo', __str_type),
        ('deployBuild', bool),
        ('permissiveChecks', bool),
        ('rms', __str_type),
        ('rmsRepo', __str_type),
        ('rmsPool', __str_type),
        ('tools', dict),
        ('toolTune', dict),
        ('package', list),
        ('packageGzipStatic', bool),
        ('packageChecksums', bool),
        ('persistent', list),
        ('writable', list),
        ('entryPoints', dict),
        ('configenv', dict),
        ('webcfg', dict),
        ('actions', dict),
        ('plugins', dict),
        ('pluginPacks', list),
    ])

    CONFIG_TUNE_VARS = OrderedDict([
        ('minMemory', 'memory'),
        ('maxMemory', 'memory'),
        ('maxTotalMemory', 'memory'),
        ('connMemory', 'memory'),
        ('connFD', int),
        ('debugOverhead', 'memory'),
        ('debugConnOverhead', 'memory'),
        ('scalable', bool),
        ('reloadable', bool),
        ('multiCore', bool),
        ('exitTimeoutMS', int),
        ('cpuWeight', 'weight'),
        ('memWeight', 'weight'),
        ('instances', int),
        ('maxInstances', int),
        ('socketTypes', list),
        ('socketProtocols', list),
        ('maxRequestSize', 'memory'),
    ])

    FUTOIN_ENV_VARS = OrderedDict([
        ('type', __str_type),
        ('persistentDir', __str_type),
        ('vars', dict),
        ('plugins', dict),
        ('pluginPacks', list),
        ('externalSetup', (bool,) + __str_type),
        ('externalServices', list),
    ])

    __slots__ = ()

    def __init__(self):
        super(ConfigMixIn, self).__init__()
        self._startup_env = dict(self._environ)
        self._overrides = None

    def _origEnv(self):
        return self._startup_env

    def __resetProcessEnv(self):
        os = self._os
        orig_env = self._startup_env
        environ = self._environ

        for (k, v) in orig_env.items():
            environ[k] = v

        to_del = set(environ.keys()) - set(orig_env.keys())

        for k in to_del:
            del environ[k]

    def __initUserHome(self):
        if 'HOME' not in self._environ:
            self._environ['HOME'] = self._ext.pwd.getpwuid(self._os.geteuid())[
                5]

    def _initConfig(self, startup=False):
        ospath = self._ospath
        os = self._os

        if not startup:
            self.__resetProcessEnv()

        errors = []

        #--
        environ = self._environ
        self.__initUserHome()
        user_home = environ['HOME']
        user_home = ospath.realpath(user_home)
        user_config_path = ospath.join(user_home, self._FUTOIN_USER_JSON)

        if not ospath.exists(user_config_path) and user_home != os.getcwd():
            user_config_path = ospath.join(user_home, self._FUTOIN_JSON)

        user_config_path = ospath.realpath(user_config_path)

        global_config_file = ospath.join(
            '/', 'etc', 'futoin', self._FUTOIN_JSON)
        global_config_file = ospath.realpath(global_config_file)

        deploy_config_file = None
        project_config_file = None
        deploy_dir = self._overrides.get('deployDir', None)

        if startup:
            pass
        elif deploy_dir:
            current = self._getDeployCurrent()
            deploy_config_file = ospath.join(deploy_dir, self._FUTOIN_JSON)
            deploy_config_file = ospath.realpath(deploy_config_file)
            project_config_file = ospath.join(
                deploy_dir, current, self._FUTOIN_JSON)
            project_config_file = ospath.realpath(project_config_file)
        else:
            project_config_file = ospath.realpath(self._FUTOIN_JSON)

        #--
        gc = {'env': {}}
        uc = {'env': {}}
        dc = {}
        pc = {}

        gc = self._pathutil.loadJSONConfig(global_config_file, gc)

        if user_config_path not in (deploy_config_file, project_config_file):
            uc = self._pathutil.loadJSONConfig(user_config_path, uc)

        if not startup:
            if project_config_file != deploy_config_file:
                pc = self._pathutil.loadJSONConfig(project_config_file, pc)

            if deploy_config_file:
                dc = self._pathutil.loadJSONConfig(deploy_config_file, dc)

        #---
        self._global_config = gc
        self._user_config = uc

        if startup:
            self._deploy_config = None
            self._project_config = None
        else:
            self._deploy_config = dc
            self._project_config = pc

        #--

        config = dict(pc)

        # Deployment config can override project config
        for (k, v) in dc.items():
            if k in ('deploy', 'env'):
                continue
            elif k == 'entryPoints':
                config_epoints = config.setdefault('entryPoints', {})

                for (ename, edef) in v.items():
                    cep = config_epoints.setdefault(ename, {})

                    for (epk, epv) in edef.items():
                        if epk == 'tune':
                            cep_tune = cep.setdefault('tune', {})
                            cep_tune.update(epv)
                        else:
                            cep[epk] = epv
            elif k == 'actions':
                config_actions = config.setdefault('actions', {})

                for (ak, av) in v.items():
                    if ak in config_actions:
                        if not isinstance(av, list):
                            lav = [av]
                        else:
                            lav = list(av)

                        try:
                            try:
                                pos = lav.index('@default')
                            except ValueError:
                                pos = lav.index('<default>')

                            cv = config_actions[ak]

                            if not isinstance(cv, list):
                                cv = [cv]

                            lav[pos:pos + 1] = cv
                            av = lav
                        except ValueError:
                            pass

                    config_actions[ak] = av
            elif k == 'persistent':
                persistent = set(config.get('persistent', []))
                persistent.update(set(v))
                config['persistent'] = list(persistent)
            else:
                config[k] = v

        self.__sanitizeConfig(config, errors)

        if 'env' in pc:
            errors.append('.env node must not be set in project config')

        if 'deploy' in pc:
            errors.append('.deploy node must not be set in project config')

        env = dict(dc.get('env', {}))

        for ct in (uc, gc):
            if 'env' not in ct or len(ct) != 1:
                errors.append(
                    'User and Global configs must have the only .env node')
                continue

            for (k, v) in ct['env'].items():
                if k == 'plugins':
                    plugins = env.setdefault('plugins', {})

                    for pk, pv in env['plugins'].items():
                        plugins.setdefault(pk, pv)
                elif k == 'pluginPacks':
                    pluginPacks = env.setdefault('pluginPacks', [])
                    pluginPacks += v
                else:
                    env.setdefault(k, v)

        self.__initEnv(env)

        config['env'] = env
        self._env = env
        config.update(self._overrides)

        if startup:
            self._config = None
        else:
            self._config = config

        self._initTools()

        #
        if not startup:
            entry_points = config.get('entryPoints', {})

            for (en, ep) in entry_points.items():
                t = self._getTool(ep['tool'])
                ep_tune = ep.setdefault('tune', {})

                for (tk, tv) in t.tuneDefaults(env).items():
                    ep_tune.setdefault(tk, tv)

            # run again to check tuneDefaults() from plugins
            self.__sanitizeEntryPoints(entry_points, errors)

        #---
        deploy = dc.get('deploy', {})
        config['deploy'] = deploy

        if '_deploy' in config:
            _deploy = config['_deploy']
            del config['_deploy']

            for (dk, dv) in _deploy.items():
                if dv is None:
                    pass
                elif dv == 'auto':
                    try:
                        del deploy[dk]
                    except KeyError:
                        pass
                else:
                    deploy[dk] = dv

        self.__sanitizeDeployConfig(config, errors)

        #---
        if not startup:
            # there is no point to export variables provided in env
            # or global configs.
            export_env = {}

            for tool in config.get('toolOrder', []):
                self._getTool(tool).exportEnv(env, export_env)

            environ = self._environ

            for (k, v) in export_env.items():
                if isinstance(v, bool):
                    v = v and '1' or ''

                try:
                    environ[k] = v
                except:
                    self._warn('Variable: {0}'.format(k))
                    raise

        #---
        if errors:
            self._errorExit(
                "Configuration issues are found:\n\n* " +
                "\n* ".join(set(errors)))

    def __sanitizeConfig(self, config, errors):
        conf_vars = self.CONFIG_VARS

        for (k, v) in config.items():
            if k not in conf_vars:
                self._warn('Removing unknown config variable "{0}"'.format(k))
                del config[k]
            elif not isinstance(v, conf_vars[k]):
                req_t = conf_vars[k]
                if isinstance(req_t, tuple):
                    req_t = req_t[0]

                errors.append(
                    'Config variable "{0}" type "{1}" is not instance of "{2}"'
                    .format(k, v.__class__.__name__, req_t.__name__)
                )

        #---
        # Make sure futoinTool is enabled, if futoin.json is present.
        # Otherwise, auto-detection gets disabled and futoin.json is not
        # updated
        tools = config.get('tools', None)

        if tools and 'futoin' not in tools:
            tools['futoin'] = True

        #---
        entry_points = config.get('entryPoints', None)

        if entry_points:
            self.__sanitizeEntryPoints(entry_points, errors)

        #---
        toolTune = config.get('toolTune', None)

        if toolTune:
            for (tn, tune) in toolTune.items():
                if not isinstance(tune, dict):
                    errors.append(
                        'Tool tune "{0}" is not of map type'.format(tn))

    def __sanitizeEntryPoints(self, entry_points, errors):
        for (en, ep) in entry_points.items():
            for k in ['tool', 'path']:
                if k not in ep:
                    errors.append(
                        'Entry point "{0}" is missing "{1}"'.format(en, k))

            if 'tune' in ep:
                ep_tune = ep['tune']

                if not isinstance(ep_tune, dict):
                    errors.append(
                        'Entry point "{0}" has invalid tune parameter'.format(en))
                    continue

                for (tk, tt) in self.CONFIG_TUNE_VARS.items():
                    try:
                        tv = ep_tune[tk]
                    except KeyError:
                        continue

                    if tt == 'memory':
                        self.__sanitizeMemory(
                            "{0}/{1}".format(en, tk), tv, errors)

                    elif tt == 'weight':
                        if not isinstance(tv, int) or tv <= 0:
                            errors.append(
                                'Weight value "{0}/{1}" must be positive'.format(en, tk))

                    elif not isinstance(tv, tt):
                        errors.append(
                            'Config tune variable "{0}" type "{1}" is not instance of "{2}"'
                            .format(tk, tv.__class__.__name__, tt.__name__)
                        )

    def __sanitizeDeployConfig(self, dc, errors):
        deploy = dc.get('deploy', {})

        if 'maxTotalMemory' in deploy:
            maxTotalMemory = deploy['maxTotalMemory']

            self.__sanitizeMemory('deploy/maxTotalMemory',
                                  maxTotalMemory, errors)

        if 'maxCpuCount' in deploy:
            val = deploy['maxCpuCount']

            if not isinstance(val, int) or val <= 0:
                errors.append(
                    '"deploy/maxCpuCount" must be a positive integer')

    def __sanitizeMemory(self, key, val, errors):
        try:
            self._configutil.parseMemory(val)
        except:
            errors.append(
                'Memory value "{0}" must be positive '
                'integer with B, K, M or G postfix (e.g. "16M")'
                .format(key))

    def __initEnv(self, env):
        env.setdefault('type', 'prod')
        env.setdefault('vars', {})
        env.setdefault('plugins', {})
        env.setdefault('pluginPacks', [])
        env.setdefault('externalSetup', False)
        env.setdefault('externalServices', [])

        timeouts = env.setdefault('timeouts', {})
        timeouts.setdefault('connect', 10)
        read_to = timeouts.setdefault('read', 60)
        timeouts.setdefault('total', read_to * 60)

        pathutil = self._pathutil
        env.setdefault('binDir', pathutil.safeJoin(
            self._environ['HOME'], 'bin'))
        pathutil.addBinPath(env['binDir'])

        #---
        env_vars = self.FUTOIN_ENV_VARS

        for (k, v) in env.items():
            if k not in env_vars:
                continue
            elif not isinstance(v, env_vars[k]):
                req_t = env_vars[k]
                if isinstance(req_t, tuple):
                    req_t = req_t[0]

                self._errorExit(
                    'Config variable "{0}" type "{1}" is not instance of "{2}"'
                    .format(k, v.__class__.__name__, req_t[0].__name__)
                )

        if env['type'] not in ('prod', 'test', 'dev'):
            self._errorExit(
                'Not valid environment type "{0}'.format(env['type']))

    def _processWcDir(self):
        ospath = self._ospath
        os = self._os

        wcDir = self._overrides['wcDir']

        try:
            os.makedirs(wcDir)
        except OSError:
            if not ospath.exists(wcDir):
                raise

        if wcDir != os.getcwd():
            self._infoLabel('Changing to: ', wcDir)
            os.chdir(wcDir)
            self._overrides['wcDir'] = os.getcwd()
            self._initConfig()
        elif self._config is None:
            self._initConfig()

    def _exportConfig(self, orig_config):
        new_config = OrderedDict()

        for cv in self.CONFIG_VARS:
            try:
                val = orig_config[cv]

                if val is not None:
                    new_config[cv] = val
            except KeyError:
                pass

        return new_config



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..subtool import SubTool


class BundlerMixIn(SubTool):
    __slots__ = ()

    def getDeps(self):
        return ['bundler']

    def _gemName(self):
        return self._name

    def _installTool(self, env):
        self._have_tool = True
        tcmd = [env['bundlerBin'], 'add', self._gemName()]
        ver = env.get(self._name + 'Ver')

        if ver:
            tcmd.append('--version={0}'.format(ver))

        try:
            self._executil.callExternal(tcmd, verbose=False)
            # self._executil.callMeaningful(tcmd)
        except Exception as e:
            self._warn(str(e))
            bundlerTools = env.setdefault('bundlerTools', {})
            bundlerTools[self._name] = ver
            return

        cmd = [env['bundlerBin'], 'install']
        self._executil.callExternal(cmd, verbose=False)
        # self._executil.callMeaningful(cmd)

    def installTool(self, env):
        if not self._have_tool:
            if self._detect.isDisabledToolsSetup(env):
                self._errorExit(
                    'Tool "{0}" must be installed externally (env config)'.format(self._name))
            else:
                self._warn(
                    'Auto-installing required "{0}" tool'.format(self._name))
                self._installTool(env)

            self.initEnv(env)

            if not self._have_tool:
                self._errorExit('Failed to install "{0}"'.format(self._name))

    def updateTool(self, env):
        # New version is installed, if not in project
        # Otherwise, the project is responsible  for updates
        self._ensureInstalled(env)

    def initEnv(self, env):
        ospath = self._ospath

        bin_path = ospath.join(env['bundlePath'], 'bin', self._name)
        env[self._name + 'Bin'] = bin_path

        if not ospath.exists('Gemfile'):
            # Fake to workaround being required outside of project root (e.g. deployment home)
            self._have_tool = True
            return

        self._have_tool = self._have_tool or ospath.exists(bin_path)

    def onRun(self, config, svc, args):
        env = config['env']
        self._ensureInstalled(env)
        self._executil.callInteractive([
            env['bundlerBin'], 'exec', self._gemName(), '--', svc['path']
        ] + args)

    def _ensureInstalled(self, env):
        bin_path = env.get(self._name + 'Bin')

        if not bin_path or not self._ospath.exists(bin_path):
            self._have_tool = False
            self.installTool(env)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..mixins.ondemand import ext as _ext

#==================


def timeouts(env, fmt):
    timeouts = env['timeouts']
    connect_to = int(timeouts['connect'])
    read_to = int(timeouts['read'])
    total_to = int(timeouts['total'])

    if fmt == 'requests':
        return (connect_to, read_to)
    elif fmt == 'curl':
        return [
            '--connect-timeout', '{0}'.format(int(connect_to)),
            '--max-time', '{0}'.format(int(total_to)),
        ]

    raise NotImplementedError('Unknown format: {0}'.format(fmt))


def requestsOptions(env, kwargs):
    kwargs['timeout'] = timeouts(env, 'requests')
    headers = kwargs.setdefault('headers', {})
    headers['User-Agent'] = 'FutoIn CID'


#==================
_memory_mult_table = {
    'B': 1,
    'K': (1 << 10),
    'M': (1 << 20),
    'G': (1 << 30),
}


def parseMemory(val):
    b = int(val[:-1])
    m = _memory_mult_table[val[-1]]

    if b <= 0:
        raise ValueError('Memory must be positive')

    return b * m


def toMemory(val):
    res = None
    old_v = 0

    for (k, v) in _memory_mult_table.items():
        if val % v:
            continue
        elif v > old_v:
            res = '{0}{1}'.format(int(val / v), k)
            old_v = v

    return res
#==================


def listify(val):
    if isinstance(val, list):
        return val

    return [val]


#==================
from collections import OrderedDict


def parseJSON(val):
    return _ext.json.loads(val, object_pairs_hook=lambda pairs: OrderedDict(pairs))

#==================


def deepMerge(target, source):
    for k, v in source.items():
        if isinstance(v, dict):
            # get node or create one
            t = target.setdefault(k, OrderedDict())
            deepMerge(t, v)
        elif isinstance(v, list):
            t = target.setdefault(k, [])
            t.extend(v)
        else:
            target[k] = v

    return target

#==================


def syslogTag(env, name):
    tag = env.get('syslogTag', None)

    if tag:
        return '{0}-{1}'.format(tag, name)

    return name



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .piptoolmixin import PipToolMixIn


class uwsgiTool(PipToolMixIn, RuntimeTool):
    """uWSGI application server container http://projects.unbit.it/uwsgi

Currently, used for Python WSGI apps.

It's possible to override uWSGI options with .tune.uwsgi parameter map.
"""
    __slots__ = ()

    def tuneDefaults(self, env):
        return {
            'minMemory': '64M',
            'connMemory': '16M',
            'debugConnOverhead': '4M',
            'connFD': 8,
            'socketTypes': ['unix', 'tcp', 'tcp6'],
            'socketType': 'unix',
            'socketProtocol': 'uwsgi',
            'scalable': True,
            'reloadable': False,  # there are too many gotchas for graceful reload
            'multiCore': False,  # make there is no uWSGI master bottleneck
            'maxRequestSize': '1M',
        }

    def onPreConfigure(self, config, runtime_dir, svc, cfg_svc_tune):
        ospath = self._ospath
        configutil = self._configutil
        env = config['env']
        deploy = config['deploy']

        name_id = '{0}-{1}'.format(svc['name'], svc['instanceId'])
        svc_tune = svc['tune']

        conf_file = 'uwsgi-{0}.ini'.format(name_id)
        conf_file = ospath.join(runtime_dir, conf_file)

        pid_file = 'uwsgi-{0}.pid'.format(name_id)
        pid_file = ospath.join(runtime_dir, pid_file)

        target_cwd = config['wcDir']

        cfg_svc_tune['uwsgiConf'] = conf_file
        cfg_svc_tune['uwsgiPid'] = pid_file

        #
        if svc_tune['socketType'] == 'unix':
            socket = svc_tune['socketPath']
        else:
            socket = '{0}:{1}'.format(
                svc_tune['socketAddress'], svc_tune['socketPort'])

        # Most of RSS is assumed to be shared with
        # uWSGI master process through CoW. So, each workes RSS
        # is minMemory=master + connMemory=worker_overhead.
        mem_limit = configutil.parseMemory(svc_tune['minMemory'])
        mem_limit += configutil.parseMemory(svc_tune['connMemory'])
        mem_limit = int(mem_limit / 1024 / 1024)
        conf = {
            'uwsgi': svc_tune.get('uwsgi', {})
        }

        uwsgi_conf = conf['uwsgi']

        uwsgi_conf.update({
            'uwsgi-socket': socket,
            'master': 1,
            'workers': svc_tune['maxConnections'],
            'strict': 1,
            'virtualenv': env['virtualenvDir'],
            'single-interpreter': 1,
            'reload-on-rss': mem_limit,
            'evil-reload-on-rss': mem_limit * 2,
            'never-swap': 1,
            'need-app': 1,
            'reaper': 1,
            'logger': 'syslog:{0}'.format(configutil.syslogTag(env, name_id)),
            'disable-logging': 1,
            'die-on-term': 1,
            'chdir': target_cwd,
            'hook-post-fork': 'chdir:{0}'.format(target_cwd),
        })

        uwsgi_conf.setdefault('max-requests', 5000)
        uwsgi_conf.setdefault('harakiri', 20)
        uwsgi_conf.setdefault('vacuum', True)

        if svc['path']:
            uwsgi_conf['wsgi-file'] = svc['path']

        self._pathutil.writeIni(conf_file, conf)

    def onRun(self, config, svc, args):
        env = config['env']
        deploy = config['deploy']
        self._executil.callInteractive([
            env['uwsgiBin'],
            svc['tune']['uwsgiConf']
        ])



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ...mixins.ondemand import ext as _ext
from .. import complex_memo as _complex_memo

from .apk import *
from .deb import *
from .ebuild import *
from .macos import *
from .pacman import *
from .rpm import *
from .java import *


def debrpm(packages):
    deb(packages)
    rpm(packages)


def generic(packages):
    apk(packages)
    deb(packages)
    brew(packages)
    pacman(packages)
    rpm(packages)

    # too different
    # emerge(packages)


@_complex_memo
def search(pattern):
    detect = _ext.detect
    pathutil = _ext.pathutil

    found = []

    #--
    apt_cache = _ext.pathutil.which('apt-cache')
    yum = _ext.pathutil.which('yum')
    dnf = _ext.pathutil.which('dnf')
    zypper = _ext.pathutil.which('zypper')

    if apt_cache:
        found = _ext.executil.callExternal(
            [apt_cache, '--names-only', 'search', pattern],
            suppress_fail=True)
        found = found.strip().split('\n')

    elif dnf or yum:
        found = _ext.executil.callExternal(
            [dnf or yum, 'search', '-q', pattern],
            suppress_fail=True)
        found = (found or '').strip().split('\n')
        found = [(f + '.').split('.')[0] for f in found]

    elif zypper:
        zypper = _ext.pathutil.which('zypper')
        res = _ext.executil.callExternal(
            [zypper, '--non-interactive', '--no-refresh',
                'search', '-t', 'package', pattern],
            suppress_fail=False)  # make user aware of failures @SLES

        found = []

        for f in (res or '').split('\n'):
            f = f.split('|')

            if len(f) > 2 and f[1] and f[1][0] != '-':
                f = f[1].strip()
                found.append(f)

    elif detect.isMacOS():
        found = _ext.install.brewSearch(pattern)

    elif detect.isAlpineLinux():
        apk = '/sbin/apk'
        found = _ext.executil.callExternal(
            [apk, 'search', pattern],
            suppress_fail=True)
        found = found.strip().split('\n')
        found = ['-'.join(f.split('-')[:-2]) for f in found]

    res = []

    for r in found:
        r = r.split()

        if r:
            r = r[0].strip()
            if r:
                res.append(r)

    return res



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .npmtoolmixin import NpmToolMixIn


class gulpTool(NpmToolMixIn, BuildTool):
    """Automate and enhance your workflow (Node.js).

Home: http://gulpjs.com/    
"""
    __slots__ = ()

    def autoDetectFiles(self):
        return 'gulpfile.js'

    def onBuild(self, config):
        cmd = [config['env']['gulpBin']]
        self._executil.callMeaningful(cmd)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool
from .piptoolmixin import PipToolMixIn


class cidTool(PipToolMixIn, BuildTool):
    "Noop FutoIn-CID - a workaround to allow CID use from virtualenv"

    __slots__ = ()

    def _pipName(self):
        return 'futoin-cid'

    def _installTool(self, env):
        source_dir = self._environ.get('CID_SOURCE_DIR', None)

        if source_dir:
            self._executil.callExternal(
                [env['pipBin'], 'install', '-q', '-e', source_dir])
        else:
            PipToolMixIn._installTool(self, env)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import importlib as _importlib
import sys
import os

__all__ = [
    'tailor_ondemand',
    'ext',
    'OnDemandMixIn',
]

#---


def tailor_ondemand(demandMap):
    class TailoredDemander(object):
        __slots__ = tuple(demandMap.keys())
        __demandMap = demandMap

        def __getattr__(self, name):
            mod = TailoredDemander.__demandMap[name]

            if isinstance(mod, tuple):
                mod, memb = mod
                mod = _importlib.import_module(mod, 'futoin.cid')
                ret = getattr(mod, memb)
            else:
                ret = _importlib.import_module(mod, 'futoin.cid')

            setattr(self, name, ret)
            return ret

    return TailoredDemander


#---
_ext_demand_map = {
    'shlex': 'shlex',
    'subprocess': 'subprocess',
    'json': 'json',
    'datetime': 'datetime',
    're': 're',
    'gzip': 'gzip',
    'shutil': 'shutil',
    'stat': 'stat',
    'time': 'time',
    'fnmatch': 'fnmatch',
    'fcntl': 'fcntl',
    'hashlib': 'hashlib',
    'signal': 'signal',
    'copy': 'copy',
    'errno': 'errno',
    'pwd': 'pwd',
    'grp': 'grp',
    'glob': 'glob',
    'dir_util': 'distutils.dir_util',
    'requests': 'requests',
    'minidom': 'xml.dom.minidom',
    'importlib': 'importlib',
    'tempfile': 'tempfile',
    'resource': 'resource',
    'binascii': 'binascii',
    'platform': 'platform',
    #
    'os': 'os',
    'ospath': 'os.path',
    'sys': 'sys',
    'detect': '.util.detect',
    'executil': '.util.executil',
    'install': '.util.install',
    'pathutil': '.util.pathutil',
    'builddep': '.util.builddep',
    'versionutil': '.util.versionutil',
    'configutil': '.util.configutil',
    'phputil': '.util.phputil',
    'github': '.util.github',
}

if sys.version_info >= (3, 0):
    _ext_demand_map['urllib'] = 'urllib.request'
else:
    _ext_demand_map['urllib'] = 'urllib2'

ext = tailor_ondemand(_ext_demand_map)()

#---
_cid_demand_map = {
    '_os': 'os',
    '_ospath': 'os.path',
    '_sys': 'sys',
    '_detect': '.util.detect',
    '_executil': '.util.executil',
    '_install': '.util.install',
    '_pathutil': '.util.pathutil',
    '_builddep': '.util.builddep',
    '_versionutil': '.util.versionutil',
    '_configutil': '.util.configutil',
    '_phputil': '.util.phputil',
}


class OnDemandMixIn(tailor_ondemand(_cid_demand_map)):
    __slots__ = ()
    _ext = ext
    _environ = os.environ



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .javatoolmixin import JavaToolMixIn


class javaTool(RuntimeTool, JavaToolMixIn):
    """Java Runtime Environment.

Home: http://openjdk.java.net/

Due to issues with Oracle's licensing, cid
supports only automatic installation of OpenJDK.

javaVer supports:
- only one digit like 6, 7, 8, 9 (Zulu JDK (http://zulu.org)
- use javaBin for custom JRE
"""
    __slots__ = ()

    def envNames(self):
        return ['javaBin', 'javaVer']

    def getVersionParts(self):
        return 1

    def _installTool(self, env):
        if env.get('javaBin', None):
            return

        detect = self._detect
        install = self._install
        ver = env['javaVer']

        if detect.isMacOS():
            install.brewTap('caskroom/versions')
            pkg_ver = 'zulu{0}'.format(ver)

            if len(install.brewSearch(pkg_ver, True)) > 0:
                install.brew(pkg_ver, True)
            else:
                install.brew('zulu', True)

            return

        if detect.isAlpineLinux():
            install.apkCommunity()
            install.apk('openjdk{0}-jre'.format(ver))
            return

        install.aptRepo(
            'zulu', 'deb http://repos.azulsystems.com/debian stable main', self._ZULU_GPG_KEY)
        install.yumRepo('zulu', 'http://repos.azulsystems.com/rhel/zulu.repo',
                        self._ZULU_GPG_KEY, releasevermax=7)
        install.zypperRepo(
            'zulu', 'http://repos.azulsystems.com/sles/latest', self._ZULU_GPG_KEY)

        # leaving here for possible future use
        # install.deb(['openjdk-{0}-jre-headless'.format(ver)])
        # install.yum(['java-1.{0}.0-openjdk'.format(ver)])
        # install.zypper(['java-1_{0}_0-openjdk'.format(ver)])
        install.debrpm(['zulu-{0}'.format(ver)])

        install.pacman(['jre{0}-openjdk'.format(ver)])
        install.emerge(['=dev-java/oracle-jre-bin-1.{0}*'.format(ver)])

    def uninstallTool(self, env):
        pass

    def initEnv(self, env):
        ospath = self._ospath
        environ = self._environ
        glob = self._ext.glob
        detect = self._detect

        if env.get('javaBin', None):
            bin_dir = ospath.dirname(env['javaBin'])
            java_home = ospath.dirname(bin_dir)

            environ['JAVA_HOME'] = java_home
            environ['JRE_HOME'] = java_home
            self._pathutil.addBinPath(bin_dir, True)

            super(javaTool, self).initEnv(env, 'java')
            return

        ver = env.setdefault('javaVer', self._LATEST_JAVA)

        candidates = [
            # Zulu
            '/usr/lib/jvm/zulu-{0}*/jre/bin/java'.format(ver),
            '/usr/lib/jvm/zulu-{0}*/bin/java'.format(ver),
            # Debian / Ubuntu
            #"/usr/lib/jvm/java-{0}-openjdk*/jre/bin/java".format(ver),
            # RedHat
            #"/usr/lib/jvm/jre-1.{0}.0/bin/java".format(ver),
            # OpenSuse
            #"/usr/lib*/jvm/jre-1.{0}.0/bin/java".format(ver),
            # Default oracle
            #"/opt/jdk/jdk1.{0}*/bin/java".format(ver),
        ]

        if detect.isGentoo() or detect.isArchLinux():
            candidates += [
                "/usr/lib/jvm/java-{0}-openjdk*/jre/bin/java".format(ver),
            ]
        elif detect.isAlpineLinux():
            candidates += [
                "/usr/lib/jvm/java-1.{0}-openjdk/jre/bin/java".format(ver),
            ]
        elif detect.isMacOS():
            candidates += [
                "/Library/Java/JavaVirtualMachines/zulu-{0}.jdk/Contents/Home/jre/bin/java".format(
                    ver),
                "/Library/Java/JavaVirtualMachines/zulu-{0}.jdk/Contents/Home/bin/java".format(
                    ver),
            ]

        for c in candidates:
            if ospath.exists(c):
                bin_name = [c]
            else:
                bin_name = glob.glob(c)

            if bin_name:
                bin_name = bin_name[0]
                bin_dir = ospath.dirname(bin_name)
                java_home = ospath.dirname(bin_dir)

                env['javaBin'] = bin_name
                environ['JAVA_HOME'] = java_home
                environ['JRE_HOME'] = java_home
                self._pathutil.addBinPath(bin_dir, True)
                self._have_tool = True
                break

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([
            env[self._name + 'Bin'], '-jar', svc['path']
        ] + args)

    _ZULU_GPG_KEY = '''
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: GnuPG v1

mQINBFNgFa8BEADTL/REB10M+TfiZOtFHqL5LHKkzTMn/O2r5iIqXGhi6iwZazFs
9S5g1eU7WMen5Xp9AREs+OvaHx91onPZ7ZiP7VpZ6ZdwWrnVk1Y/HfI59tWxmNYW
DmKYBGMj4EUpFPSE9EnFj7dm1WdlCvpognCwZQl9D3BseGqN7OLHfwqqmOlbYN9h
HYkT+CaqOoWDIGMB3UkBlMr0GuujEP8N1gxg7EOcSCsZH5aKtXubdUlVSphfAAwD
z4MviB39J22sPBnKmaOT3TUTO5vGeKtC9BAvtgA82jY2TtCEjetnfK/qtzj/6j2N
xVUbHQydwNQVRU92A7334YvCbn3xUUNI0WOscdmfpgCU0Z9Gb2IqDb9cMjgUi8F6
MG/QY9/CZjX62XrHRPm3aXsCJOVh/PO1sl2A/rvv8AkpJKYyhm6T8OBFptCsA3V4
Oic7ZyYhqV0u2r4NON+1MoUeuuoeY2tIrbRxe3ffVOxPzrESzSbc8LC2tYaP+wGd
W0f57/CoDkUzlvpReCUI1Bv5zP4/jhC63Rh6lffvSf2tQLwOsf5ivPhUtwUfOQjg
v9P8Wc8K7XZpSOMnDZuDe9wuvB/DiH/P5yiTs2RGsbDdRh5iPfwbtf2+IX6h2lNZ
XiDKt9Gc26uzeJRx/c7+sLunxq6DLIYvrsEipVI9frHIHV6fFTmqMJY6SwARAQAB
tEdBenVsIFN5c3RlbXMsIEluYy4gKFBhY2thZ2Ugc2lnbmluZyBrZXkuKSA8cGtp
LXNpZ25pbmdAYXp1bHN5c3RlbXMuY29tPokCOAQTAQIAIgUCU2AVrwIbAwYLCQgH
AwIGFQgCCQoLBBYCAwECHgECF4AACgkQsZmDYSGb2cnJ8xAAz1V1PJnfOyaRIP2N
Ho2uRwGdPsA4eFMXb4Z08eGjDMD3b9WW3D0XnCLbJpaZ6klz0W0s2tcYSneTBaSs
RAqxgJgBZ5ZMXtrrHld/5qFoBbStLZLefmcPhnfvamwHDCTLUex8NIAI1u3e9Rhb
5fbH+gpuYpwHX7hz0FOfpn1sxR03UyxU+ey4AdKe9LG3TJVnB0WcgxpobpbqweLH
yzcEQCNoFV3r1rlE13Y0aE31/9apoEwiYvqAzEmE38TukDLl/Qg8rkR1t0/lok2P
G6pWqdN7pmoUovBTvDi5YOthcjZcdOTXXn2Yw4RZVF9uhRsVfku1Eg25SnOje3uY
smtQLME4eESbePdjyV/okCIle66uHZse+7gNyNmWpf01hM+VmAySIAyKa0Ku8AXZ
MydEcJTebrNfW9uMLsBx3Ts7z/CBfRng6F8louJGlZtlSwddTkZVcb26T20xeo0a
ZvdFXM2djTi/a5nbBoZQL85AEeV7HaphFLdPrgmMtS8sSZUEVvdaxp7WJsVuF9cO
Nxsvx40OYTvfco0W41Lm8/sEuQ7YueEVpZxiv5kX56GTU9vXaOOi+8Z7Ee2w6Adz
4hrGZkzztggs4tM9geNYnd0XCdZ/ICAskKJABg7biDD1PhEBrqCIqSE3U497vibQ
Mpkkl/Zpp0BirhGWNyTg8K4JrsQ=
=d320
-----END PGP PUBLIC KEY BLOCK-----
'''

    def tuneDefaults(self, env):
        return {
            'minMemory': '64M',
            'debugOverhead': '128M',
            'connMemory': '100K',
            'debugConnOverhead': '1M',
            'socketType': 'tcp',
            'scalable': False,
            'reloadable': False,
            'multiCore': True,
        }



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..rmstool import RmsTool


class artifactoryTool(RmsTool):
    """JFrog Artifactory: Artifact Repository Manager.

Home: https://www.jfrog.com/artifactory/

Note 1: JFrog CLI is used for operation. So, you have all flexibility
to setup different authentication schemes:
> cte jfrog rt config --enc-password=false
or
> cte jfrog rt config --apiKey=
The first found matching URL will be used.

Alternatively, you can set env in ~/.futoin.json. Please avoid
setting env variables for security reasons.

Note 2: Artifactory OSS API is limited and may not support all commands
    (e.g. pool create). So, Artifactory Pro is assumed.

Note 3: only the part before '/' of 'rms_pool' becomes actual RMS pool
"""
    __slots__ = ()

    def getDeps(self):
        return ['jfrog']

    def envNames(self):
        return ['artifactoryUser', 'artifactoryPassword', 'artifactoryApiKey']

    def initEnv(self, env):
        self._environ['JFROG_CLI_LOG_LEVEL'] = 'ERROR'
        self._have_tool = True

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath
        self._checkFileUpload(config, rms_pool, package_list)

        server_cfg = self._getServerConfig(config)

        for package in package_list:
            self._executil.callExternal([
                config['env']['jfrogBin'],
                'rt', 'upload',
                '--server-id={0}'.format(server_cfg['serverId']),
                '--url={0}'.format(config['rmsRepo']),
                '--flat=true',
                '--recursive=false',
                '--regexp=false',
                package,
                '{0}/{1}'.format(rms_pool, ospath.basename(package))
            ])

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        self._checkFileUpload(config, dst_pool, package_list)

        server_cfg = self._getServerConfig(config)

        for package in package_list:
            self._executil.callExternal([
                config['env']['jfrogBin'],
                'rt', 'copy',
                '--server-id={0}'.format(server_cfg['serverId']),
                '--url={0}'.format(config['rmsRepo']),
                '--flat=true',
                '--recursive=false',
                '{0}/{1}'.format(src_pool, package),
                '{0}/{1}'.format(dst_pool, package)
            ])

    def rmsGetList(self, config, rms_pool, package_hint):
        ospath = self._ospath
        (pool, path) = self._prepParts(rms_pool, package_hint or '*')

        result = self._callArtifactory(
            config,
            'GET', '/api/search/pattern',
            params={
                'pattern': '{0}:{1}'.format(pool, path),
            }
        )

        result.raise_for_status()

        result = result.json()
        result = [ospath.basename(r) for r in result['files']]
        return result

    def rmsRetrieve(self, config, rms_pool, package_list):
        server_cfg = self._getServerConfig(config)

        for package in package_list:
            self._executil.callExternal([
                config['env']['jfrogBin'],
                'rt', 'download',
                '--server-id={0}'.format(server_cfg['serverId']),
                '--url={0}'.format(config['rmsRepo']),
                '--flat=true',
                '--recursive=false',
                '{0}/{1}'.format(rms_pool, package),
                package])

    def rmsPoolCreate(self, config, rms_pool):
        rms_pool = rms_pool.split('/')[0]
        check = self._callArtifactory(
            config,
            'GET', '/api/repositories/{0}'.format(rms_pool),
        )

        if check.ok:
            check = check.json()

            if check['packageType'] != 'generic':
                self._errorExit('Artifactory pool {0} already exists with different type {1}'.format(
                    rms_pool, check['packageType']))

            return

        result = self._callArtifactory(
            config,
            'PUT', '/api/repositories/{0}'.format(rms_pool),
            json={
                'rclass': 'local',
                'packageType': 'generic',
                'repoLayoutRef': 'simple-default',
            }
        )
        result.raise_for_status()

    def rmsPoolList(self, config):
        result = self._callArtifactory(
            config,
            'GET', '/api/repositories'
        )

        result.raise_for_status()

        result = result.json()
        result = [r['key'] for r in result]
        return result

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        result = self._callArtifactory(
            config,
            'GET', '/api/storage/{0}/{1}'.format(rms_pool, package)
        )

        result.raise_for_status()
        result = result.json()

        if hash_type == 'sha256' and 'sha256' not in result['checksums']:
            (pool, path) = self._prepParts(rms_pool, package)

            result = self._callArtifactory(
                config,
                'POST', '/api/checksum/sha256',
                json={
                    'repoKey': pool,
                    'path': path,
                }
            )
            result.raise_for_status()

            result = self._callArtifactory(
                config,
                'GET', '/api/storage/{0}/{1}'.format(rms_pool, package)
            )

            result.raise_for_status()
            result = result.json()

        return result['checksums'][hash_type]

    def _callArtifactory(self, config, method, path, **kwargs):
        import requests

        rms_repo = config['rmsRepo']
        if rms_repo[-1] == '/':
            path = path[1:]

        url = rms_repo + path

        server_cfg = self._getServerConfig(config)

        if 'apiKey' in server_cfg:
            headers = kwargs.setdefault('headers', {})
            headers['X-JFrog-Art-Api'] = server_cfg['apiKey']
        elif 'password' in server_cfg:
            kwargs['auth'] = (server_cfg['user'], server_cfg['password'])

        self._configutil.requestsOptions(config['env'], kwargs)

        self._info('HTTP call {0} {1}'.format(method, url))
        return requests.request(method, url, **kwargs)

    def _getServerConfig(self, config, repeat=False):
        url = config['rmsRepo']

        if url[-1] != '/':
            url += '/'

        pathutil = self._pathutil
        jfrog_cfg = pathutil.safeJoin(
            self._environ['HOME'], '.jfrog', 'jfrog-cli.conf')
        jfrog_cfg = pathutil.loadJSONConfig(jfrog_cfg)

        if jfrog_cfg:
            for server_cfg in jfrog_cfg.get('artifactory', []):
                if 'serverId' in server_cfg and server_cfg.get('url', None) == url:
                    return server_cfg

        env = config['env']

        if repeat:
            pass
        elif 'artifactoryUser' in env and 'artifactoryPassword' in env:
            self._executil.callExternal([
                env['jfrogBin'],
                'rt', 'config',
                '--interactive=false',
                '--enc-password=false',
                '--url={0}'.format(url),
                '--user={0}'.format(env['artifactoryUser']),
                '--password={0}'.format(env['artifactoryPassword']),
                self._genServerId(),
            ])
            return self._getServerConfig(config, True)
        elif 'artifactoryApiKey' in env:
            self._executil.callExternal([
                env['jfrogBin'],
                'rt', 'config',
                '--interactive=false',
                '--url={0}'.format(url),
                '--apikey={0}'.format(env['artifactoryApiKey']),
                self._genServerId(),
            ])
            return self._getServerConfig(config, True)

        self._errorExit(
            'Please make sure Artifactory server is configured in JFrog CLI. See tool desc.')

    def _genServerId(self):
        rndm = self._ext.binascii.hexlify(self._os.urandom(8))
        rndm = self._executil.toString(rndm)
        return 'cid-' + rndm

    def _checkFileUpload(self, config, rms_pool, package_list):
        ospath = self._ospath

        for package in package_list:
            package = ospath.basename(package)
            res = self.rmsGetList(config, rms_pool, package)

            if res:
                self._errorExit(
                    'Package {0} already exists in RMS pool {1}'.format(package, rms_pool))

    def _prepParts(self, rms_pool, package):
        rms_pool = rms_pool.split('/', 1)

        path = ''

        if len(rms_pool) == 2:
            path += rms_pool[1] + '/'

        path += package

        return (rms_pool[0], path)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runtimetool import RuntimeTool
from .bashtoolmixin import BashToolMixIn


class goTool(BashToolMixIn, RuntimeTool):
    """The Go Programming Language

Home: https://golang.org/

All versions are installed through GVM.

Only binary releases of Golang are supported for installation
through CID, but you can install source releases through
"cte gvm install sourcetag".
"""
    __slots__ = ()

    def getDeps(self):
        return ['gvm', 'bash', 'binutils', 'gcc']

    def _installTool(self, env):
        if self._detect.isAlpineLinux():
            self._install.apkCommunity()
            self._install.apk('go')
            return

        # in case GVM is already installed without these deps
        self._install.deb(['bison', 'build-essential'])
        self._install.rpm(['bison', 'glibc-devel'])
        self._install.emergeDepsOnly(['dev-lang/go'])
        self._install.pacman(['bison', 'glibc', ])
        self._install.apk('bison')
        self._builddep.essential()

        self._callBash(env,
                       'source {0} && gvm install {1} --binary'
                       .format(env['gvmInit'], env['goVer'])
                       )

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        self._callBash(env,
                       'source {0} && gvm uninstall {1}'
                       .format(env['gvmInit'], env['goVer'])
                       )
        self._have_tool = False

    def envNames(self):
        return ['goVer', 'goBin']

    def initEnv(self, env):
        if self._detect.isAlpineLinux():
            super(goTool, self).initEnv(env)
            return

        if not env.get('goVer', None):
            try:
                cmd = 'source {0} && gvm listall'.format(env['gvmInit'])
                ver_list = self._callBash(env, cmd, verbose=False)
                ver_list = ver_list.split("\n")

                rex = self._ext.re.compile('^go[0-9]+\.[0-9]+(\.[0-9]+)?$')

                ver_list = [v.strip() for v in ver_list]
                ver_list = filter(lambda x: x and rex.match(x), ver_list)

                env['goVer'] = self._versionutil.latest(list(ver_list))
            except Exception as e:
                self._warn(str(e))
                return

        ver = env['goVer']

        try:
            env_to_set = self._callBash(env,
                                        'source {0} && \
                gvm use {1} >/dev/null && \
                env | egrep -i "(gvm|golang)"'.format(env['gvmInit'], ver),
                                        verbose=False
                                        )
        except:
            return

        if env_to_set:
            self._pathutil.updateEnvFromOutput(env_to_set)
            super(goTool, self).initEnv(env)

    def onRun(self, config, svc, args):
        env = config['env']
        self._executil.callInteractive([
            env[self._name + 'Bin'], 'run', svc['path']
        ] + args)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..buildtool import BuildTool


class cmakeTool(BuildTool):
    """Build, Test and Package Your Software With CMake.

Home: https://cmake.org/

CMake creates a build folder and does all processing in it.
Build folder is configurable through cmakeBuildDir env.

On release tagging, CMakeLists.txt the following replacements are done:
1. 'VERSION ".*" # AUTO-REPLACE' -> new version
"""
    __slots__ = ()

    def getOrder(self):
        return -10

    def autoDetectFiles(self):
        return 'CMakeLists.txt'

    def envNames(self):
        return ['cmakeBin', 'cmakeBuildDir']

    def _installTool(self, env):
        self._install.deb(['build-essential'])
        self._install.rpm(['gcc', 'gcc-c++'])
        self._install.pacman(['gcc'])
        self._builddep.essential()

        self._install.debrpm(['cmake'])
        self._install.emerge(['dev-util/cmake'])
        self._install.pacman(['cmake'])
        self._install.apk(['cmake'])
        self._install.brew('cmake')

    def initEnv(self, env):
        env.setdefault('cmakeBuildDir', 'build')
        super(cmakeTool, self).initEnv(env)

    def onPrepare(self, config):
        ospath = self._ospath
        os = self._os
        build_dir = config['env']['cmakeBuildDir']
        self._pathutil.rmTree(build_dir)

        os.mkdir(build_dir)
        cmd = [config['env']['cmakeBin'], '-H' +
               config['wcDir'], '-B' + build_dir]
        self._executil.callMeaningful(cmd)

    def onBuild(self, config):
        ospath = self._ospath
        os = self._os
        build_dir = config['env']['cmakeBuildDir']

        if ospath.exists(build_dir):
            cmd = [config['env']['cmakeBin'], build_dir]
            self._executil.callMeaningful(cmd)
        else:
            self.onPrepare(config)

        cmd = [config['env']['cmakeBin'], '--build', build_dir]
        self._executil.callMeaningful(cmd)

    def updateProjectConfig(self, config, updates):
        re = self._ext.re

        def cmake_updater(content):
            if 'version' in updates:
                return re.sub(
                    r'VERSION.*".*".*# AUTO-REPLACE',
                    'VERSION "{0}" # AUTO-REPLACE'.format(updates['version']),
                    content,
                    flags=re.MULTILINE
                )

        return self._pathutil.updateTextFile(
            self.autoDetectFiles(), cmake_updater)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .data import DataSlots


class LockMixIn(DataSlots):
    __slots__ = ()

    __DEPLOY_LOCK_FILE = '.futoin-deploy.lock'
    __MASTER_LOCK_FILE = '.futoin-master.lock'
    __GLOBAL_LOCK_FILE = '.futoin-global.lock'

    def __init__(self):
        super(LockMixIn, self).__init__()
        self._deploy_lock = None
        self._master_lock = None
        self._global_lock = None

    def __lockCommon(self, lock, file, flags):
        fcntl = self._ext.fcntl
        os = self._os

        assert getattr(self, lock) is None

        lockfd = os.open(file, os.O_WRONLY | os.O_CREAT)
        setattr(self, lock, lockfd)

        try:
            fcntl.flock(lockfd, flags)
        except Exception as e:
            self._errorExit('FAILED to acquire{0}: {1}'.format(
                lock.replace('_', ' '), e))

    def __unlockCommon(self, lock):
        fcntl = self._ext.fcntl
        lockfd = getattr(self, lock)
        fcntl.flock(lockfd, fcntl.LOCK_UN)
        self._os.close(lockfd)
        setattr(self, lock, None)

    def _deployLock(self):
        fcntl = self._ext.fcntl
        self.__lockCommon(
            '_deploy_lock',
            self.__deployLockFile(),
            fcntl.LOCK_EX | fcntl.LOCK_NB
        )

    def _deployUnlock(self):
        self.__unlockCommon('_deploy_lock')

    def _requireDeployLock(self):
        if self._deploy_lock is None:
            self._errorExit('Deploy lock must be already acquired')

    def _globalLock(self):
        fcntl = self._ext.fcntl
        self.__lockCommon(
            '_global_lock',
            self.__globalLockFile(),
            fcntl.LOCK_EX
        )

    def _globalUnlock(self):
        self.__unlockCommon('_global_lock')

    def _masterLock(self):
        fcntl = self._ext.fcntl
        self.__lockCommon(
            '_master_lock',
            self.__masterLockFile(),
            fcntl.LOCK_EX | fcntl.LOCK_NB
        )

    def _masterUnlock(self):
        self.__unlockCommon('_master_lock')

    def __deployLockFile(self):
        return self._pathutil.safeJoin(self._overrides['deployDir'], self.__DEPLOY_LOCK_FILE)

    def __masterLockFile(self):
        return self._pathutil.safeJoin(self._overrides['deployDir'], self.__MASTER_LOCK_FILE)

    def __globalLockFile(self):
        return self._pathutil.safeJoin(self._environ['HOME'], self.__GLOBAL_LOCK_FILE)

    def _checkDeployLock(self):
        ospath = self._ospath
        placeholder = self.__deployLockFile()
        deploy_dir = self._overrides['deployDir']

        if not ospath.exists(placeholder) and self._os.listdir(deploy_dir):
            self._errorExit(
                "Deployment dir '{0}' is missing safety placeholder {1}."
                .format(deploy_dir, ospath.basename(placeholder))
            )



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..migrationtool import MigrationTool


class liquibaseTool(MigrationTool):
    """Liquibase - source control for your database.

Home: http://www.liquibase.org/

If LIQUIBASE_HOME is set, the tool must be installed manually.
If liquibaseVer is used, it must be set to full version (GitHub API limitation).

Liquibase migrate is run only if liquibase.properties file is present.
"changeLogFile" must be set in liquibase.properties.

Required drivers can be controlled through liquibaseDrivers variable.
Currently supported:
* mysql
* postgresql
* sqlite
* mssql

Others, are more JVM-specific and most likely will be bundled with JVM app.
"""
    __slots__ = ()

    LIQUIBASE_PROPERTIES = 'liquibase.properties'

    def autoDetectFiles(self):
        return self.LIQUIBASE_PROPERTIES

    def getDeps(self):
        return ['java', 'tar', 'gzip']

    def getVersionParts(self):
        return 3

    def envNames(self):
        return ['liquibaseVer', 'liquibaseDir', 'liquibaseDrivers']

    def _installTool(self, env):
        ospath = self._ospath
        os = self._os
        github = self._ext.github

        lb_dir = env['liquibaseDir']
        ver = env.get('liquibaseVer', 'latest')

        if ver == 'latest':
            release = ver
        else:
            release = 'tags/liquibase-parent-{0}'.format(ver)

        info = github.releaseInfo(env, 'liquibase/liquibase', release)

        found_ver = info['name'].replace('v', '')
        dst = ospath.join(lb_dir, found_ver)
        asset = github.findAsset(info['assets'], 'application/x-gzip')

        if not ospath.exists(dst):
            self._pathutil.downloadExtract(
                env, asset['browser_download_url'],
                dst, 'z')

        if ver == 'latest':
            latest = ospath.join(lb_dir, 'latest')
            try:
                os.unlink(latest)
            except OSError:
                pass
            os.symlink(found_ver, latest)

    def _updateTool(self, env):
        self._installTool(env)

    def uninstallTool(self, env):
        ver = env.get('liquibaseVer', 'latest')
        inst_dir = env['liquibaseDir']

        if ver != 'latest':
            inst_dir = self._pathutil.safeJoin(inst_dir, ver)

        self._pathutil.rmTree(inst_dir)

    def initEnv(self, env):
        ospath = self._ospath

        #---
        inst_dir = self._environ.get('LIQUIBASE_HOME', None)
        fail_on_missing = False

        if inst_dir:
            fail_on_missing = True
        else:
            lb_dir = ospath.join(
                self._pathutil.userHome(),
                '.local', 'liquibasebin')
            lb_dir = env.setdefault('liquibaseDir', lb_dir)
            ver = env.get('liquibaseVer', 'latest')
            inst_dir = ospath.join(lb_dir, ver)

        lb_bin = ospath.join(inst_dir, 'liquibase')

        if ospath.exists(lb_bin):
            self._install.ensureJDBC(
                env,
                ospath.join(inst_dir, 'lib'),
                env.get('liquibaseDrivers', '').split())
            env['liquibaseBin'] = lb_bin
            self._environ['LIQUIBASE_HOME'] = inst_dir
            self._have_tool = True
        elif fail_on_missing:
            self._errorExit('Unset LIQUIBASE_HOME or '
                            'setup the tool manually.')

    def onMigrate(self, config):
        if self._ospath.exists(self.LIQUIBASE_PROPERTIES):
            self.onExec(config['env'], ['update'], replace=False)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..rmstool import RmsTool


class archivaTool(RmsTool):
    """Apache Archiva: The Build Artifact Repository Manager.
Home: https://archiva.apache.org/

NOT IMPLEMENTED YET!
"""
    __slots__ = ()

    def envNames(self):
        return ['archivaUser', 'archivaPassword']

    def initEnv(self, env):
        self._have_tool = True

    def rmsUpload(self, config, rms_pool, package_list):
        ospath = self._ospath

        for package in package_list:
            package_basename = ospath.basename(package)
            path = '/repository/{0}/{1}'.format(rms_pool, package_basename)

            res = self._callArchiva(config, 'HEAD', path)

            if res.ok:
                self._errorExit(
                    'Package {0} already exists on RMS'.format(package_basename))

            hashes = self.rmsCalcHashes(package)

            with open(package, 'rb') as pf:
                res = self._callArchiva(
                    config,
                    'PUT', path,
                    data=pf
                )

                res.raise_for_status()

                for (hash_type, hash_value) in hashes.items():
                    res = self._callArchiva(
                        config,
                        'PUT', '{0}.{1}'.format(path, hash_type),
                        data='{0}  {1}'.format(hash_value, package_basename),
                    )
                    res.raise_for_status()

    def rmsPromote(self, config, src_pool, dst_pool, package_list):
        for package in package_list:
            src_path = '/repository/{0}/{1}'.format(src_pool, package)
            dst_path = '/repository/{0}/{1}'.format(dst_pool, package)

            res = self._callArchiva(
                config,
                'COPY', src_path,
                headers={
                    'Depth': '0',
                    'Overwrite': 'F',
                    'Destination': dst_path,
                }
            )
            res.raise_for_status()

            for hash_type in self.ALLOWED_HASH_TYPES:
                src_hash = '{0}.{1}'.format(src_path, hash_type)

                res = self._callArchiva(config, 'GET', src_hash)

                if res.status_code == 404:
                    continue

                res.raise_for_status()

                res = self._callArchiva(
                    config,
                    'PUT', '{0}.{1}'.format(dst_path, hash_type),
                    data=res.text
                )
                res.raise_for_status()

    def rmsGetList(self, config, rms_pool, package_hint):
        ospath = self._ospath

        apires = self._callArchiva(
            config,
            'PROPFIND', '/repository/{0}/'.format(rms_pool),
            headers={
                'Depth': '1',
                'Content-type': 'text/xml; charset="utf-8"'
            },
            data='''<?xml version="1.0" encoding="utf-8" ?>
<propfind xmlns="DAV:">
  <prop>
    <displayname/>
    <iscollection/>
  </prop>
</propfind>'''
        )
        apires.raise_for_status()

        minidom = self._ext.minidom
        try:
            apidom = minidom.parseString(apires.text)
        except:
            self._warn(apires.text)
            raise

        ret = []

        for response in apidom.getElementsByTagNameNS('*', 'response'):
            if response.getElementsByTagNameNS('*', 'iscollection')[0].firstChild.nodeValue == '0':
                name = response.getElementsByTagNameNS(
                    '*', 'displayname')[0].firstChild.nodeValue
                (sname, sext) = ospath.splitext(name)

                if sext and sext[1:] in self.ALLOWED_HASH_TYPES:
                    continue

                ret.append(name)

        return ret

    def rmsRetrieve(self, config, rms_pool, package_list):
        shutil = self._ext.shutil

        for package in package_list:
            result = self._callArchiva(
                config,
                'GET',
                '/repository/{0}/{1}'.format(rms_pool, package),
                stream=True
            )
            result.raise_for_status()

            with open(package, 'wb') as f:
                result.raw.decode_content = True
                shutil.copyfileobj(result.raw, f)

    def rmsPoolCreate(self, config, rms_pool):
        res = self._callArchiva(
            config,
            'GET', '/restServices/archivaServices/managedRepositoriesService/getManagedRepository/{0}'.format(
                rms_pool),
            headers={'Accept': 'application/json'},
        )

        if res.ok and len(res.text) and res.json()['id'] == rms_pool:
            return

        res = self._callArchiva(
            config,
            'POST', '/restServices/archivaServices/managedRepositoriesService/addManagedRepository',
            headers={'Accept': 'application/json'},
            json={
                'id': rms_pool,
                'name': rms_pool,
                'blockRedeployments': True,
                'cronExpression': "0 0 * * * ?",
                'indexDirectory': '.indexer',
                'layout': "default",
                'location': rms_pool,
                'releases': True,
                'scanned': False,  # produces issue with Lucene locks
                'skipPackedIndexCreation': True,
                'snapshots': False,
            }
        )

        if not res.ok:
            self._warn(res.text)
        res.raise_for_status()
        assert res.json()['id'] == rms_pool

    def rmsPoolList(self, config):
        res = self._callArchiva(
            config,
            'GET', '/restServices/archivaServices/managedRepositoriesService/getManagedRepositories',
            headers={'Accept': 'application/json'},
        )
        res.raise_for_status()
        res = res.json()

        return [r['id'] for r in res]

    def rmsGetHash(self, config, rms_pool, package, hash_type):
        path = '{0}/{1}.{2}'.format(rms_pool, package, hash_type)

        res = self._callArchiva(
            config,
            'GET', '/repository/{0}/{1}.{2}'.format(
                rms_pool, package, hash_type)
        )
        res.raise_for_status()
        return res.text.split()[0].strip()

    def _callArchiva(self, config, method, path, **kwargs):
        env = config['env']

        rms_repo = config['rmsRepo']
        if rms_repo[-1] == '/':
            path = path[1:]

        url = rms_repo + path

        if 'archivaUser' in env and 'archivaPassword' in env:
            kwargs['auth'] = (env['archivaUser'], env['archivaPassword'])

        self._configutil.requestsOptions(env, kwargs)

        self._info('HTTP call {0} {1}'.format(method, url))
        return self._ext.requests.request(method, url, **kwargs)



#
# Copyright 2015-2018 Andrey Galkin <andrey@futoin.org>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..runenvtool import RunEnvTool
from .bashtoolmixin import BashToolMixIn


class nvmTool(BashToolMixIn, RunEnvTool):
    """Node Version Manager.

Home: https://github.com/creationix/nvm    
"""
    __slots__ = ()

    NVM_LATEST = '$(git describe --abbrev=0 --tags --match "v[0-9]*" origin/master)'

    def getDeps(self):
        return ['bash', 'git']

    def _installTool(self, env):
        nvm_dir = env['nvmDir']
        nvm_git = env.get('nvmGit', 'https://github.com/creationix/nvm.git')
        nvm_ver = env.get('nvmVer', self.NVM_LATEST)

        self._callBash(env,
                       'git clone {1} {0}; \
               cd {0} && git fetch && git reset --hard && git checkout {2}'
                       .format(nvm_dir, nvm_git, nvm_ver))

    def _updateTool(self, env):
        nvm_dir = env['nvmDir']
        nvm_ver = env.get('nvmVer', self.NVM_LATEST)

        self._callBash(env,
                       'cd {0} && git fetch && git reset --hard && git checkout {1}'
                       .format(nvm_dir, nvm_ver))

    def uninstallTool(self, env):
        nvm_dir = env['nvmDir']
        self._pathutil.rmTree(nvm_dir)
        self._have_tool = False

    def envNames(self):
        return ['nvmDir', 'nvmGit', 'nvmVer']

    def initEnv(self, env):
        ospath = self._ospath
        nvm_dir = ospath.join(self._environ['HOME'], '.nvm')
        nvm_dir = env.setdefault('nvmDir', nvm_dir)
        env_init = ospath.join(nvm_dir, 'nvm.sh')
        env['nvmInit'] = env_init
        self._have_tool = ospath.exists(env_init)

    def onExec(self, env, args, replace=True):
        cmd = '. {0} && nvm {1}'.format(
            env['nvmInit'], self._ext.subprocess.list2cmdline(args))
        self._callBashInteractive(env, cmd, replace=replace)
